# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for
# the Earth and Planetary Sciences
# Copyright (C) 2012 - 2021 by the BurnMan team, released under the GNU
# GPL v2 or later.


"""
Introducing BurnMan |version|
=============================

Overview
--------

BurnMan is an open source mineral physics and seismological toolkit written in Python
which can enable a user to calculate (or fit) the physical and chemical properties
of endmember minerals, fluids/melts, solid solutions, and composite assemblages.

Properties which BurnMan can calculate include:

  - the thermodynamic free energies, allowing phase equilibrium calculations,
    endmember activities, chemical potentials and oxygen (and other) fugacities.

  - entropy, enabling the user to calculate isentropes for a given assemblage.

  - volume, to allow the user to create density profiles.

  - seismic velocities, including Voigt-Reuss-Hill and Hashin-Strikman bounds
    and averages.

The toolkit itself comes with a large set of classes and functions which are
designed to allow the user to easily combine mineral physics with
geophysics, and geodynamics. The features of BurnMan include:

  - the full codebase, which includes implementations of many static and thermal equations of state
    (including Vinet, Birch Murnaghan, Mie-Debye-Grueneisen, Modified Tait),
    and solution models (ideal, symmetric, asymmetric, subregular).
  - popular endmember and solution datasets already coded into burnman-usable format
    (including :cite:`HP2011`, :cite:`Stixrude2005` and :cite:`Stixrude2011`)
  - Optimal least squares fitting routines for multivariate data with (potentially correlated) errors
    in pressure and temperature. As an example, such functions can be used to
    simultaneously fit volumes, seismic velocities and enthalpies.
  - a "Planet" class, which self-consistently calculates gravity profiles, mass, moment of
    inertia of planets given the chemical and temperature structure of a planet
  - published geotherms
  - a tutorial on the basic use of BurnMan
  - a large collection of annotated examples
  - a set of high-level functions which create files readable by seismological and geodynamic software,
    including: Mineos :cite:`Masters2011`, AxiSEM :cite:`NissenMeyer2014` and ASPECT
  - an extensive suite of unit tests to ensure code functions as intended
  - a series of benchmarks comparing BurnMan output with published data
  - a directory containing user-contributed code from published papers

BurnMan makes extensive use of `SciPy <http://www.scipy.org/>`_,
`NumPy <http://www.numpy.org/>`_ and `SymPy <http://www.sympy.org/>`_
which are widely used Python libraries for scientific computation.
`Matplotlib <http://matplotlib.org/>`_ is used
to display results and produce publication quality figures.
The computations are consistently formulated in terms of SI units.

The code documentation including class and function descriptions can be found online at
http://burnman.readthedocs.io.

This software has been designed to allow the end-user a great deal of freedom
to do whatever calculations they may wish and to add their own modules.
The underlying Python classes have been designed to make new endmember,
solid solution and composite models easy to read and create.
We have endeavoured to provide examples and benchmarks which cover the
most popular uses of the software, some of which are included in the figure below.
This list is certainly not exhaustive, and we will definitely have missed interesting
applications. We will be very happy to accept contributions in
form of corrections, examples, or new features.

Structure
---------
.. image:: figures/structure.png


.. _ref-installation:

Installation
------------

Requirements
^^^^^^^^^^^^

  - Python 3.7+
  - Python modules: NumPy, SciPy, SymPy, Matplotlib


Source code
^^^^^^^^^^^
The source code can be found at https://github.com/geodynamics/burnman.

Install under Ubuntu
^^^^^^^^^^^^^^^^^^^^

1. Install dependencies using apt by opening a terminal window and entering
   ``sudo apt-get install python python-scipy python-numpy python-sympy python-matplotlib git``
2. Clone the BurnMan repository ``git clone https://github.com/geodynamics/burnman.git``
3. Go to the Burnman examples directory and type:
   ``python example_beginner.py``
   Figures should show up, indicating that it is working.


Install on a Mac
^^^^^^^^^^^^^^^^

1. get Xcode
2. If you don't have Python yet, download it (for free) from
   python.org/download . Make sure to use Python 3.7+.
   To check your version of python, type the following in a
   terminal: ``python --version``
3. Install the latest Numpy version from http://sourceforge.net/projects/numpy/files/NumPy/
4. Install the latest Scipy from http://sourceforge.net/projects/scipy/files/
5. Install the latest Sympy from http://sourceforge.net/projects/sympy/files/
6. Install the latest Matplotlib from http://sourceforge.net/projects/matplotlib/files/matplotlib/matplotlib-1.1.1/
7. Clone the BurnMan repository ``git clone https://github.com/geodynamics/burnman.git``
8. Go to the Burnman examples directory and type ``python example_beginner.py``
   Figures should show up, indicating that it is working.

Install under Windows
^^^^^^^^^^^^^^^^^^^^^

To get Python running under Windows:

1. Download Python from http://www.python.org/ and install
2. Go to http://www.lfd.uci.edu/~gohlke/pythonlibs/#numpy, download and install
3. Go to http://www.lfd.uci.edu/~gohlke/pythonlibs/#scipy, download and install
4. Go to http://www.lfd.uci.edu/~gohlke/pythonlibs/#sympy, download and install
5. Go to http://www.lfd.uci.edu/~gohlke/pythonlibs/#matplotlib, download and install
6. Download BurnMan from github (https://github.com/geodynamics/burnman)
7. Open Python Shell (IDLE Python GUI)
8. File -- Open -- find one of the example files
9. Run the module (or press F5)


Citing BurnMan
--------------

If you use BurnMan in your work, we ask that you cite the following publications:

  - Myhill, R., Cottaar, S., Heister, T., Rose, I., and Unterborn, C. (2021):
    BurnMan v1.0.0 [Software]. Computational Infrastructure for Geodynamics. Zenodo.
    `(https://doi.org/10.5281/zenodo.5552443) <https://doi.org/10.5281/zenodo.5552443>`_

  - Cottaar S., Heister, T., Rose, I., and Unterborn, C., 2014, BurnMan: A
    lower mantle mineral physics toolkit, Geochemistry, Geophysics, and
    Geosystems, 15(4), 1164-1179 `(https://doi.org/10.1002/2013GC005122)
    <https://doi.org/10.1002/2013GC005122>`_

Contributing to BurnMan
-----------------------

If you would like to contribute bug fixes, new functions or new modules
to the existing codebase, please contact us at info@burnman.org or make a
pull request at `https://github.com/geodynamics/burnman <https://github.com/geodynamics/burnman>`_.

BurnMan also includes a contrib directory that contains python and ipython
scripts used to reproduce published results. We welcome the submission of
new contributions to this directory. As with the contribution of code,
please contact us at info@burnman.org or make a pull request at
`https://github.com/geodynamics/burnman <https://github.com/geodynamics/burnman>`_.

Acknowledgement and Support
---------------------------

  - This project was initiated at, and follow-up research support was received
    through, Cooperative Institute of Deep Earth Research, CIDER (NSF FESD
    grant 1135452) -- see `www.deep-earth.org <http://www.deep-earth.org>`_

  - We thank all the members of the CIDER Mg/Si team for their input:
    Valentina Magni, Yu Huang, JiaChao Liu, Marc Hirschmann, and Barbara
    Romanowicz. We also thank Lars Stixrude for providing benchmarking calculations
    and Zack Geballe, Motohiko Murakami, Bill McDonough, Quentin Williams,
    Wendy Panero, and Wolfgang Bangerth for helpful discussions.

  - We thank CIG (`www.geodynamics.org <http://www.geodynamics.org>`_) for support
    and accepting our donation of BurnMan as an official project.

"""
from __future__ import absolute_import
from .version import version as __version__

# Classes and associated functions for representing rocks and minerals:
from .classes.material import Material, material_property
from .classes.perplex import PerplexMaterial
from .classes.mineral import Mineral
from .classes.combinedmineral import CombinedMineral
from .classes.solutionmodel import SolutionModel
from .classes.solidsolution import SolidSolution
from .classes.composite import Composite
from .classes.anisotropy import AnisotropicMaterial
from .classes.anisotropicmineral import AnisotropicMineral
from .classes.anisotropicmineral import cell_parameters_to_vectors
from .classes.anisotropicmineral import cell_vectors_to_parameters
from .classes.mineral_helpers import HelperLowHighPressureRockTransition
from .classes.mineral_helpers import HelperSpinTransition
from .classes.mineral_helpers import HelperRockSwitcher

# Other classes
from .classes.composition import Composition
from .classes.layer import Layer
from .classes.planet import Planet
from .classes.polytope import MaterialPolytope
from .classes import seismic
from .classes import averaging_schemes

# Mineral library
from . import minerals

# Equations of state
from . import eos

# Tools
from . import tools
from .tools import geotherm
from .tools.equilibration import equilibrate
from .tools.partitioning import calculate_nakajima_fp_pv_partition_coefficient

# Optimization functions
from .optimize import composition_fitting
from .optimize import linear_fitting
from .optimize import nonlinear_fitting
from .optimize import nonlinear_solvers
from .optimize import eos_fitting



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for
# the Earth and Planetary Sciences
# Copyright (C) 2012 - 2021 by the BurnMan team, released under the GNU
# GPL v2 or later.

"""
RS_2014_liquids
^^^^^^^^^^^^^^^

Liquids from Ramo and Stixrude (2014) FPMD simulations.
There are some typos in the article which have been corrected where marked
with the help of David Munoz Ramo.
"""

from burnman import constants
from ..classes.mineral import Mineral
from ..classes.solidsolution import SolidSolution
from ..classes.solutionmodel import *

from ..tools.chemistry import dictionarize_formula, formula_mass

class Fe2SiO4_liquid(Mineral):
    def __init__(self):
        formula = 'Fe2SiO4'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'Fe2SiO4_liquid',
            'formula': formula,
            'equation_of_state': 'dks_l',
            'V_0':  59.7717e-6, # modified for T_0
            'T_0':  1900., # corrected
            'O_theta': 1 ,
            'O_f': 4 ,
            'm': 0.6 ,
            'a': np.array([[-4252948.0, 997810.188],
                           [-599315.125, 12032.8936],
                           [12572739., 7299239.5],
                           [53442800.0, -26791676.0],
                           [52981912.0, 0.]]), # corrected order
            'zeta_0': 0.0161350928 , # 0.0166734, # the comment is a refit to David's dataset
            'xi': 0.34431043 , # 0.34431053, # the comment is a refit to David's dataset
            'Tel_0': 1919.3553 , # 1921.6813, # the comment is a refit to David's dataset
            'eta': 0.0127067110 , # 0.0127067, # the comment is a refit to David's dataset
            'spin_a': [-0.00011134, 0.00010863],
            'spin_b': [3.53793, -3.81421, 2.83703, -0.676241],
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}
        Mineral.__init__(self)



from __future__ import absolute_import
# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for
# the Earth and Planetary Sciences
# Copyright (C) 2012 - 2021 by the BurnMan team, released under the GNU
# GPL v2 or later.


import numpy as np
import warnings

from . import modified_tait as mt
from . import murnaghan as murn
from . import equation_of_state as eos

from . import einstein


class HP_TMT(eos.EquationOfState):

    """
    Base class for the thermal equation of state based on
    the generic modified Tait equation of state (class MT),
    as described in :cite:`HP2011`.


    An instance "m" of a Mineral can be assigned this
    equation of state with the command m.set_method('hp_tmt')
    (or by initialising the class with the param
    equation_of_state = 'hp_tmt'
    """

    def volume(self, pressure, temperature, params):
        """
        Returns volume [m^3] as a function of pressure [Pa] and temperature [K]
        EQ 12
        """
        Pth = self.__relative_thermal_pressure(temperature, params)
        return mt.volume(pressure - Pth, params)

    def pressure(self, temperature, volume, params):
        """
        Returns pressure [Pa] as a function of temperature [K] and volume[m^3]
        EQ B7
        """
        Pth = self.__relative_thermal_pressure(temperature, params)
        return mt.modified_tait(params['V_0'] / volume, params) + Pth

    def grueneisen_parameter(self, pressure, temperature, volume, params):
        """
        Returns grueneisen parameter [unitless] as a function of pressure,
        temperature, and volume.
        """
        alpha = self.thermal_expansivity(
            pressure, temperature, volume, params)
        K_T = self.isothermal_bulk_modulus(
            pressure, temperature, volume, params)
        C_V = self.molar_heat_capacity_v(pressure, temperature, volume, params)
        return alpha * K_T * volume / C_V

    def isothermal_bulk_modulus(self, pressure, temperature, volume, params):
        """
        Returns isothermal bulk modulus [Pa] as a function of pressure [Pa],
        temperature [K], and volume [m^3].  EQ 13+2
        """
        Pth = self.__relative_thermal_pressure(temperature, params)
        return mt.bulk_modulus(pressure - Pth, params)

    # calculate the shear modulus as a function of P, V, and T
    def shear_modulus(self, pressure, temperature, volume, params):
        """
        Not implemented.
        Returns 0.
        Could potentially apply a fixed Poissons ratio as a rough estimate.
        """
        return 0.

    # Cv, heat capacity at constant volume
    def molar_heat_capacity_v(self, pressure, temperature, volume, params):
        """
        Returns heat capacity at constant volume at the pressure, temperature,
        and volume [J/K/mol].
        """
        C_p = self.molar_heat_capacity_p(pressure, temperature, volume, params)
        V = self.volume(pressure, temperature, params)
        alpha = self.thermal_expansivity(pressure, temperature, volume, params)
        K_T = self.isothermal_bulk_modulus(
            pressure, temperature, volume, params)
        return C_p - V * temperature * alpha * alpha * K_T

    def thermal_expansivity(self, pressure, temperature, volume, params):
        """
        Returns thermal expansivity at the pressure, temperature,
        and volume [1/K]. This function replaces -Pth in EQ 13+1
        with P-Pth for non-ambient temperature
        """
        a, b, c = mt.tait_constants(params)
        Pth = self.__relative_thermal_pressure(temperature, params)
        psubpth = pressure - params['P_0'] - Pth

        C_V0 = einstein.molar_heat_capacity_v(
            params['T_0'], params['T_einstein'], params['n'])
        C_V = einstein.molar_heat_capacity_v(
            temperature, params['T_einstein'], params['n'])
        alpha = params['a_0'] * (C_V / C_V0) * 1. / (
            (1. + b * psubpth) * (a + (1. - a) * np.power((1 + b * psubpth), c)))
        return alpha

    def molar_heat_capacity_p0(self, temperature, params):
        """
        Returns heat capacity at ambient pressure as a function of temperature
        [J/K/mol]. Cp = a + bT + cT^-2 + dT^-0.5 in :cite:`HP2011`.
        """
        Cp = params['Cp'][0] + params['Cp'][1] * temperature + params['Cp'][2] * \
            np.power(temperature, -2.) + params[
                'Cp'][3] * np.power(temperature, -0.5)
        return Cp

    def molar_heat_capacity_p_einstein(self, pressure, temperature, volume, params):
        """
        Returns heat capacity at constant pressure at the pressure,
        temperature, and volume, using the C_v and Einstein model [J/K/mol]
        WARNING: Only for comparison with internally self-consistent C_p
        """
        alpha = self.thermal_expansivity(pressure, temperature, volume, params)
        gr = self.grueneisen_parameter(pressure, temperature, volume, params)
        C_v = self.molar_heat_capacity_v(pressure, temperature, volume, params)
        C_p = C_v * (1. + gr * alpha * temperature)
        return C_p

    def adiabatic_bulk_modulus(self, pressure, temperature, volume, params):
        """
        Returns adiabatic bulk modulus [Pa] as a function of pressure [Pa],
        temperature [K], and volume [m^3].
        """
        K_T = self.isothermal_bulk_modulus(
            pressure, temperature, volume, params)
        C_p = self.molar_heat_capacity_p(pressure, temperature, volume, params)
        C_v = self.molar_heat_capacity_v(pressure, temperature, volume, params)
        K_S = K_T * C_p / C_v
        return K_S

    def gibbs_free_energy(self, pressure, temperature, volume, params):
        """
        Returns the gibbs free energy [J/mol] as a function of pressure [Pa]
        and temperature [K].
        """
        # Calculate temperature and pressure integrals
        a, b, c = mt.tait_constants(params)
        Pth = self.__relative_thermal_pressure(temperature, params)

        psubpth = pressure - params['P_0'] - Pth

        # EQ 13
        if pressure != params['P_0']:
            intVdP = (pressure - params['P_0']) * params['V_0'] * (
                1. - a + (a * (np.power((1. - b * Pth), 1. - c) - np.power((1. + b * (psubpth)), 1. - c)) / (b * (c - 1.) * (pressure - params['P_0']))))
        else:
            intVdP = 0.
        return params['H_0'] + self.__intCpdT(temperature, params) - temperature * (params['S_0'] + self.__intCpoverTdT(temperature, params)) + intVdP

    def helmholtz_free_energy(self, pressure, temperature, volume, params):
        return self.gibbs_free_energy(pressure, temperature, volume, params) - pressure * self.volume(pressure, temperature, params)

    def entropy(self, pressure, temperature, volume, params):
        """
        Returns the entropy [J/K/mol] as a function of pressure [Pa]
        and temperature [K].
        """
        a, b, c = mt.tait_constants(params)
        Pth = self.__relative_thermal_pressure(temperature, params)

        ksi_over_ksi_0 = einstein.molar_heat_capacity_v(temperature, params['T_einstein'], params[
                                                  'n']) / einstein.molar_heat_capacity_v(params['T_0'], params['T_einstein'], params['n'])

        dintVdpdT = (params['V_0'] * params['a_0'] * params['K_0'] * a * ksi_over_ksi_0) * (
            np.power((1. + b * (pressure - params['P_0'] - Pth)), 0. - c) - np.power((1. - b * Pth), 0. - c))
        return params['S_0'] + self.__intCpoverTdT(temperature, params) + dintVdpdT

    def enthalpy(self, pressure, temperature, volume, params):
        """
        Returns the enthalpy [J/mol] as a function of pressure [Pa]
        and temperature [K].
        """
        gibbs = self.gibbs_free_energy(pressure, temperature, volume, params)
        entropy = self.entropy(pressure, temperature, volume, params)
        return gibbs + temperature * entropy

    def molar_heat_capacity_p(self, pressure, temperature, volume, params):
        """
        Returns the heat capacity [J/K/mol] as a function of pressure [Pa]
        and temperature [K].
        """
        a, b, c = mt.tait_constants(params)
        T = temperature
        T_e = params['T_einstein']
        n = params['n']
        Pth = self.__relative_thermal_pressure(T, params)

        ksi_over_ksi_0 = einstein.molar_heat_capacity_v(T, T_e, n) \
                         / einstein.molar_heat_capacity_v(params['T_0'], T_e, n)

        dintVdpdT = (params['V_0'] * params['a_0'] * params['K_0'] * a * ksi_over_ksi_0) * (
            np.power((1. + b * (pressure - params['P_0'] - Pth)), 0. - c) - np.power((1. - b * Pth), 0. - c))

        dSdT0 = params['V_0'] * params['K_0'] * np.power((ksi_over_ksi_0 * params['a_0']), 2.0) * \
                (np.power((1. + b * (pressure - params['P_0'] - Pth)), -1. - c)
                 - np.power((1. + b * (-Pth)), -1. - c))

        x = T_e/T
        dCv_einstdT = -(einstein.molar_heat_capacity_v(T, T_e, n)
                        * (1 - 2./x + 2./(np.exp(x) - 1.)) * x/T)

        dSdT1 = -dintVdpdT * dCv_einstdT \
                / einstein.molar_heat_capacity_v(T, T_e, n)

        dSdT = dSdT0 + dSdT1
        return self.molar_heat_capacity_p0(temperature, params) + temperature * dSdT

    def __thermal_pressure(self, T, params):
        """
        Returns thermal pressure [Pa] as a function of T [K]
        EQ 12 - 1 of :cite:`HP2011`.
        """

        # This is basically the mie-gruneisen equation of state for thermal
        # pressure using an Einstein model for heat capacity.  The additional
        # assumption that they make is that alpha*K/Cv, (or gamma / V) is
        # constant over a wide range of compressions.

        # Note that the xi function in HP2011 is just the Einstein heat capacity
        # divided by 3nR. This function is *not* used to calculate the
        # heat capacity - Holland and Powell (2011) prefer the additional
        # freedom provided by their polynomial expression.

        E_th = einstein.thermal_energy(T, params['T_einstein'], params['n'])
        C_V0 = einstein.molar_heat_capacity_v(
            params['T_0'], params['T_einstein'], params['n'])
        P_th = params['a_0'] * params['K_0'] / C_V0 * E_th
        return P_th

    def __relative_thermal_pressure(self, T, params):
        """
        Returns relative thermal pressure [Pa] as a function of T-params['T_0'] [K]
        EQ 12 - 1 of :cite:`HP2011`.
        """
        return self.__thermal_pressure(T, params) - \
            self.__thermal_pressure(params['T_0'], params)

    def __intCpdT(self, temperature, params):
        """
        Returns the thermal addition to the standard state enthalpy [J/mol]
        at ambient pressure [Pa]
        """
        return (params['Cp'][0] * temperature + 0.5 * params['Cp'][1] * np.power(temperature, 2.) - params['Cp'][2] / temperature + 2. * params['Cp'][3] * np.sqrt(temperature)) - (params['Cp'][0] * params['T_0'] + 0.5 * params['Cp'][1] * params['T_0'] * params['T_0'] - params['Cp'][2] / params['T_0'] + 2.0 * params['Cp'][3] * np.sqrt(params['T_0']))

    def __intCpoverTdT(self, temperature, params):
        """
        Returns the thermal addition to the standard state entropy [J/K/mol]
        at ambient pressure [Pa]
        """
        return (params['Cp'][0] * np.log(temperature) + params['Cp'][1] * temperature - 0.5 * params['Cp'][2] / np.power(temperature, 2.) - 2.0 * params['Cp'][3] / np.sqrt(temperature)) - (params['Cp'][0] * np.log(params['T_0']) + params['Cp'][1] * params['T_0'] - 0.5 * params['Cp'][2] / (params['T_0'] * params['T_0']) - 2.0 * params['Cp'][3] / np.sqrt(params['T_0']))

    def validate_parameters(self, params):
        """
        Check for existence and validity of the parameters
        """
        if 'T_0' not in params:
            params['T_0'] = 298.15

        # If standard state enthalpy and entropy are not included
        # this is presumably deliberate, as we can model density
        # and bulk modulus just fine without them.
        # Just add them to the dictionary as nans.
        if 'H_0' not in params:
            params['H_0'] = float('nan')
        if 'S_0' not in params:
            params['S_0'] = float('nan')

        # First, let's check the EoS parameters for Tref
        mt.MT.validate_parameters(mt.MT(), params)

        # Now check all the required keys for the
        # thermal part of the EoS are in the dictionary
        expected_keys = ['H_0', 'S_0', 'V_0', 'Cp', 'a_0', 'n', 'molar_mass']
        for k in expected_keys:
            if k not in params:
                raise KeyError('params object missing parameter : ' + k)

        # The following line estimates the Einstein temperature
        # according to the empirical equation of
        # Holland and Powell, 2011; base of p.346, para.1
        if 'T_einstein' not in params:
            params['T_einstein'] = 10636. / \
                (params['S_0'] / params['n'] + 6.44)

        # Finally, check that the values are reasonable.
        if params['T_0'] < 0.:
            warnings.warn('Unusual value for T_0', stacklevel=2)
        if params['G_0'] is not float('nan') and (params['G_0'] < 0. or params['G_0'] > 1.e13):
            warnings.warn('Unusual value for G_0', stacklevel=2)
        if params['Gprime_0'] is not float('nan') and (params['Gprime_0'] < -5. or params['Gprime_0'] > 10.):
            warnings.warn('Unusual value for Gprime_0', stacklevel=2)

        # no test for H_0
        if params['S_0'] is not float('nan') and params['S_0'] < 0.:
            warnings.warn('Unusual value for S_0', stacklevel=2)
        if params['V_0'] < 1.e-7 or params['V_0'] > 1.e-2:
            warnings.warn('Unusual value for V_0', stacklevel=2)

        if self.molar_heat_capacity_p0(params['T_0'], params) < 0.:
            warnings.warn('Negative heat capacity at T_0', stacklevel=2)
        if self.molar_heat_capacity_p0(2000., params) < 0.:
            warnings.warn('Negative heat capacity at 2000K', stacklevel=2)

        if params['a_0'] < 0. or params['a_0'] > 1.e-3:
            warnings.warn('Unusual value for a_0', stacklevel=2)

        if params['n'] < 1. or params['n'] > 1000.:
            warnings.warn('Unusual value for n', stacklevel=2)
        if params['molar_mass'] < 0.001 or params['molar_mass'] > 10.:
            warnings.warn('Unusual value for molar_mass', stacklevel=2)


class HP_TMTL(eos.EquationOfState):

    """
    Base class for the thermal equation of state
    described in :cite:`HP1998`, but with the Modified Tait as the static part,
    as described in :cite:`HP2011`.

    An instance "m" of a Mineral can be assigned this
    equation of state with the command m.set_method('hp_tmtL')
    (or by initialising the class with the param
    equation_of_state = 'hp_tmtL'
    """

    def _V_T_1bar(self, temperature, params):
        # Constant thermal expansivity at standard state pressure
        # (p.348 of HP2011)
        return params['V_0']*np.exp(params['a_0']*(temperature - params['T_0']))

    def _K_T_1bar(self, temperature, params):
        # Linear bulk modulus dependence as in HP1998 (p.348 of HP2011)
        return params['K_0'] + params['dKdT_0']*(temperature - params['T_0'])

    def volume(self, pressure, temperature, params):
        """
        Returns volume [m^3] as a function of pressure [Pa] and temperature [K]
        """
        self.static_params['V_0'] = self._V_T_1bar(temperature, params)
        self.static_params['K_0'] = self._K_T_1bar(temperature, params)
        return mt.volume(pressure, self.static_params)

    def pressure(self, temperature, volume, params):
        """
        Returns pressure [Pa] as a function of temperature [K] and volume[m^3]
        """
        self.static_params['V_0'] = self._V_T_1bar(temperature, params)
        self.static_params['K_0'] = self._K_T_1bar(temperature, params)
        return mt.pressure(volume, self.static_params)

    def grueneisen_parameter(self, pressure, temperature, volume, params):
        """
        Returns grueneisen parameter [unitless] as a function of pressure,
        temperature, and volume.
        """
        alpha = self.thermal_expansivity(pressure, temperature, volume, params)
        K_T = self.isothermal_bulk_modulus(pressure, temperature,
                                           volume, params)
        C_V = self.molar_heat_capacity_v(pressure, temperature,
                                         volume, params)
        return alpha * K_T * volume / C_V

    def isothermal_bulk_modulus(self, pressure, temperature, volume, params):
        """
        Returns isothermal bulk modulus [Pa] as a function of pressure [Pa],
        temperature [K], and volume [m^3].
        """
        self.static_params['V_0'] = self._V_T_1bar(temperature, params)
        self.static_params['K_0'] = self._K_T_1bar(temperature, params)
        return mt.bulk_modulus(pressure, self.static_params)

    # calculate the shear modulus as a function of P, V, and T
    def shear_modulus(self, pressure, temperature, volume, params):
        """
        Not implemented.
        Returns 0.
        Could potentially apply a fixed Poissons ratio as a rough estimate.
        """
        return 0.

    # Cv, heat capacity at constant volume
    def molar_heat_capacity_v(self, pressure, temperature, volume, params):
        """
        Returns heat capacity at constant volume at the pressure, temperature,
        and volume [J/K/mol].
        """
        C_p = self.molar_heat_capacity_p(pressure, temperature, volume, params)
        V = self.volume(pressure, temperature, params)
        alpha = self.thermal_expansivity(pressure, temperature, volume, params)
        K_T = self.isothermal_bulk_modulus(pressure, temperature,
                                           volume, params)
        return C_p - V * temperature * alpha * alpha * K_T

    def thermal_expansivity(self, pressure, temperature, volume, params):
        """
        Returns thermal expansivity at the pressure, temperature,
        and volume [1/K]
        """
        # The derivation of the high pressure thermal expansivity is tedious,
        # so here we take a numerical derivative.
        # TODO Derive and use the analytical derivative.
        dT = 0.1
        self.static_params['V_0'] = self._V_T_1bar(temperature+dT/2., params)
        self.static_params['K_0'] = self._K_T_1bar(temperature+dT/2., params)
        volume1 = mt.volume(pressure, self.static_params)
        self.static_params['V_0'] = self._V_T_1bar(temperature-dT/2., params)
        self.static_params['K_0'] = self._K_T_1bar(temperature-dT/2., params)
        volume0 = mt.volume(pressure, self.static_params)

        return 2.*(volume1 - volume0)/(volume1 + volume0)/dT

    def molar_heat_capacity_p0(self, temperature, params):
        """
        Returns heat capacity at ambient pressure as a function of temperature
        [J/K/mol]
        Cp = a + bT + cT^-2 + dT^-0.5 in :cite:`HP1998`.
        """
        Cp = (params['Cp'][0]
              + params['Cp'][1] * temperature
              + params['Cp'][2] * np.power(temperature, -2.)
              + params['Cp'][3] * np.power(temperature, -0.5))
        return Cp

    def adiabatic_bulk_modulus(self, pressure, temperature, volume, params):
        """
        Returns adiabatic bulk modulus [Pa] as a function of pressure [Pa],
        temperature [K], and volume [m^3].
        """
        K_T = self.isothermal_bulk_modulus(pressure, temperature,
                                           volume, params)
        C_p = self.molar_heat_capacity_p(pressure, temperature, volume, params)
        C_v = self.molar_heat_capacity_v(pressure, temperature, volume, params)
        K_S = K_T * C_p / C_v
        return K_S

    def gibbs_free_energy(self, pressure, temperature, volume, params):
        """
        Returns the gibbs free energy [J/mol] as a function of pressure [Pa]
        and temperature [K].
        """
        self.static_params['V_0'] = self._V_T_1bar(temperature, params)
        self.static_params['K_0'] = self._K_T_1bar(temperature, params)
        return (params['H_0'] + self.__intCpdT(temperature, params)
                - temperature * (params['S_0']
                                 + self.__intCpoverTdT(temperature, params))
                + mt.intVdP(pressure, self.static_params))

    def helmholtz_free_energy(self, pressure, temperature, volume, params):
        return (self.gibbs_free_energy(pressure, temperature, volume, params)
                - pressure * self.volume(pressure, temperature, params))

    def entropy(self, pressure, temperature, volume, params):
        """
        Returns the entropy [J/K/mol] as a function of pressure [Pa]
        and temperature [K].
        """
        # The derivation of the entropy is tedious,
        # so here we take a numerical derivative.
        # TODO Derive and use the analytical derivative.
        dT = 0.1
        G1 = self.gibbs_free_energy(pressure, temperature+dT/2., volume, params)
        G0 = self.gibbs_free_energy(pressure, temperature-dT/2., volume, params)

        return (G0 - G1)/dT

    def enthalpy(self, pressure, temperature, volume, params):
        """
        Returns the enthalpy [J/mol] as a function of pressure [Pa]
        and temperature [K].
        """
        gibbs = self.gibbs_free_energy(pressure, temperature, volume, params)
        entropy = self.entropy(pressure, temperature, volume, params)
        return gibbs + temperature * entropy

    def molar_heat_capacity_p(self, pressure, temperature, volume, params):
        """
        Returns the heat capacity [J/K/mol] as a function of pressure [Pa]
        and temperature [K].
        """
        # The differentiation is tedious, so for now we just take the
        # numerical derivative of S
        # TODO Derive and use the analytical derivative.
        dT = 0.1
        S1 = self.entropy(pressure, temperature+dT/2., volume, params)
        S0 = self.entropy(pressure, temperature-dT/2., volume, params)
        return temperature * (S1 - S0)/dT

    def __intCpdT(self, temperature, params):
        """
        Returns the thermal addition to the standard state enthalpy [J/mol]
        at ambient pressure [Pa]
        """
        return ((params['Cp'][0] * temperature
                 + 0.5 * params['Cp'][1] * np.power(temperature, 2.)
                 - params['Cp'][2] / temperature
                 + 2. * params['Cp'][3] * np.sqrt(temperature))
                - (params['Cp'][0] * params['T_0']
                   + 0.5 * params['Cp'][1] * params['T_0'] * params['T_0']
                   - params['Cp'][2] / params['T_0']
                   + 2.0 * params['Cp'][3] * np.sqrt(params['T_0'])))

    def __intCpoverTdT(self, temperature, params):
        """
        Returns the thermal addition to the standard state entropy [J/K/mol]
        at ambient pressure [Pa]
        """
        return ((params['Cp'][0] * np.log(temperature)
                 + params['Cp'][1] * temperature
                 - 0.5 * params['Cp'][2] / np.power(temperature, 2.)
                 - 2.0 * params['Cp'][3] / np.sqrt(temperature))
                - (params['Cp'][0] * np.log(params['T_0'])
                   + params['Cp'][1] * params['T_0']
                   - 0.5 * params['Cp'][2] / (params['T_0'] * params['T_0'])
                   - 2.0 * params['Cp'][3] / np.sqrt(params['T_0'])))

    def validate_parameters(self, params):
        """
        Check for existence and validity of the parameters
        """
        if 'T_0' not in params:
            params['T_0'] = 298.15

        # If standard state enthalpy and entropy are not included
        # this is presumably deliberate, as we can model density
        # and bulk modulus just fine without them.
        # Just add them to the dictionary as nans.
        if 'H_0' not in params:
            params['H_0'] = float('nan')
        if 'S_0' not in params:
            params['S_0'] = float('nan')

        # First, let's check the EoS parameters for Tref
        mt.MT.validate_parameters(mt.MT(), params)
        self.static_params = {'V_0': params['V_0'],
                              'K_0': params['K_0'],
                              'Kprime_0': params['Kprime_0'],
                              'Kdprime_0': params['Kdprime_0'],
                              'P_0': params['P_0']}

        # Now check all the required keys for the
        # thermal part of the EoS are in the dictionary
        expected_keys = ['H_0', 'S_0', 'V_0', 'Cp', 'a_0', 'dKdT_0',
                         'n', 'molar_mass']
        for k in expected_keys:
            if k not in params:
                raise KeyError('params object missing parameter : ' + k)

        # Finally, check that the values are reasonable.
        if params['T_0'] < 0.:
            warnings.warn('Unusual value for T_0', stacklevel=2)
        if ((params['G_0'] is not float('nan')
             and (params['G_0'] < 0. or params['G_0'] > 1.e13))):
            warnings.warn('Unusual value for G_0', stacklevel=2)
        if ((params['Gprime_0'] is not float('nan')
             and (params['Gprime_0'] < -5. or params['Gprime_0'] > 10.))):
            warnings.warn('Unusual value for Gprime_0', stacklevel=2)

        # no test for H_0 or S_0 (several HP endmembers have S_0 < 0)
        if params['V_0'] < 1.e-7 or params['V_0'] > 1.e-2:
            warnings.warn('Unusual value for V_0', stacklevel=2)

        if self.molar_heat_capacity_p0(params['T_0'], params) < 0.:
            warnings.warn('Negative heat capacity at T_0', stacklevel=2)
        if self.molar_heat_capacity_p0(2000., params) < 0.:
            warnings.warn('Negative heat capacity at 2000K', stacklevel=2)

        if params['a_0'] < 0. or params['a_0'] > 1.e-3:
            warnings.warn('Unusual value for a_0', stacklevel=2)

        if params['n'] < 1. or params['n'] > 1000.:
            warnings.warn('Unusual value for n', stacklevel=2)
        if params['molar_mass'] < 0.001 or params['molar_mass'] > 10.:
            warnings.warn('Unusual value for molar_mass', stacklevel=2)


class HP98(eos.EquationOfState):

    """
    Base class for the thermal equation of state
    described in :cite:`HP1998`.

    An instance "m" of a Mineral can be assigned this
    equation of state with the command m.set_method('hp98')
    (or by initialising the class with the param
    equation_of_state = 'hp98'
    """

    def _V_T_1bar(self, temperature, params):
        return params['V_0']*(1. + params['a_0']*(temperature - params['T_0'])
                              - 20.*params['a_0']*(np.sqrt(temperature)
                                                   - np.sqrt(params['T_0'])))

    def _K_T_1bar(self, temperature, params):
        return params['K_0'] + params['dKdT_0']*(temperature - params['T_0'])

    def volume(self, pressure, temperature, params):
        """
        Returns volume [m^3] as a function of pressure [Pa] and temperature [K]
        """
        return murn.volume(pressure,
                           self._V_T_1bar(temperature, params),
                           self._K_T_1bar(temperature, params),
                           params['Kprime_0'])

    def pressure(self, temperature, volume, params):
        """
        Returns pressure [Pa] as a function of temperature [K] and volume[m^3]
        """
        return murn.pressure(volume,
                             self._V_T_1bar(temperature, params),
                             self._K_T_1bar(temperature, params),
                             params['Kprime_0'])

    def grueneisen_parameter(self, pressure, temperature, volume, params):
        """
        Returns grueneisen parameter [unitless] as a function of pressure,
        temperature, and volume.
        """
        alpha = self.thermal_expansivity(pressure, temperature, volume, params)
        K_T = self.isothermal_bulk_modulus(pressure, temperature,
                                           volume, params)
        C_V = self.molar_heat_capacity_v(pressure, temperature,
                                         volume, params)
        return alpha * K_T * volume / C_V

    def isothermal_bulk_modulus(self, pressure, temperature, volume, params):
        """
        Returns isothermal bulk modulus [Pa] as a function of pressure [Pa],
        temperature [K], and volume [m^3].
        """
        return murn.bulk_modulus(pressure,
                                 self._K_T_1bar(temperature, params),
                                 params['Kprime_0'])

    # calculate the shear modulus as a function of P, V, and T
    def shear_modulus(self, pressure, temperature, volume, params):
        """
        Not implemented.
        Returns 0.
        Could potentially apply a fixed Poissons ratio as a rough estimate.
        """
        return 0.

    # Cv, heat capacity at constant volume
    def molar_heat_capacity_v(self, pressure, temperature, volume, params):
        """
        Returns heat capacity at constant volume at the pressure, temperature,
        and volume [J/K/mol].
        """
        C_p = self.molar_heat_capacity_p(pressure, temperature, volume, params)
        V = self.volume(pressure, temperature, params)
        alpha = self.thermal_expansivity(pressure, temperature, volume, params)
        K_T = self.isothermal_bulk_modulus(pressure, temperature,
                                           volume, params)
        return C_p - V * temperature * alpha * alpha * K_T

    def thermal_expansivity(self, pressure, temperature, volume, params):
        """
        Returns thermal expansivity at the pressure, temperature,
        and volume [1/K]
        """
        VT = self._V_T_1bar(temperature, params)
        KT = self._K_T_1bar(temperature, params)
        volume = murn.volume(pressure, VT, KT, params['Kprime_0'])
        dVTdT = params['V_0']*params['a_0']*(1. - 10./np.sqrt(temperature))
        g = volume / VT
        dgdKT = (pressure
                 * np.power(1. + pressure*params['Kprime_0']/KT,
                            -1-1./params['Kprime_0'])
                 / (KT*KT))
        dVdT = dVTdT*g + VT*dgdKT*params['dKdT_0']
        return dVdT / volume

    def molar_heat_capacity_p0(self, temperature, params):
        """
        Returns heat capacity at ambient pressure as a function of temperature
        [J/K/mol]
        Cp = a + bT + cT^-2 + dT^-0.5 in :cite:`HP1998`.
        """
        Cp = (params['Cp'][0]
              + params['Cp'][1] * temperature
              + params['Cp'][2] * np.power(temperature, -2.)
              + params['Cp'][3] * np.power(temperature, -0.5))
        return Cp

    def adiabatic_bulk_modulus(self, pressure, temperature, volume, params):
        """
        Returns adiabatic bulk modulus [Pa] as a function of pressure [Pa],
        temperature [K], and volume [m^3].
        """
        K_T = self.isothermal_bulk_modulus(pressure, temperature,
                                           volume, params)
        C_p = self.molar_heat_capacity_p(pressure, temperature, volume, params)
        C_v = self.molar_heat_capacity_v(pressure, temperature, volume, params)
        K_S = K_T * C_p / C_v
        return K_S

    def gibbs_free_energy(self, pressure, temperature, volume, params):
        """
        Returns the gibbs free energy [J/mol] as a function of pressure [Pa]
        and temperature [K].
        """
        return (params['H_0'] + self.__intCpdT(temperature, params)
                - temperature * (params['S_0']
                                 + self.__intCpoverTdT(temperature, params))
                + murn.intVdP(pressure,
                              self._V_T_1bar(temperature, params),
                              self._K_T_1bar(temperature, params),
                              params['Kprime_0']))

    def helmholtz_free_energy(self, pressure, temperature, volume, params):
        return (self.gibbs_free_energy(pressure, temperature, volume, params)
                - pressure * self.volume(pressure, temperature, params))

    def entropy(self, pressure, temperature, volume, params):
        """
        Returns the entropy [J/K/mol] as a function of pressure [Pa]
        and temperature [K].
        """
        # The entropy involves differentiating intdVdp
        # with respect to temperature.
        # We do this using the product and chain rules

        VT = self._V_T_1bar(temperature, params)
        KT = self._K_T_1bar(temperature, params)
        dVTdT = params['V_0']*params['a_0']*(1. - 10./np.sqrt(temperature))
        g = murn.intVdP(pressure, VT, KT, params['Kprime_0']) / VT
        dgdKT = (((pressure/KT + 1.)
                 * np.power(1. + pressure*params['Kprime_0']/KT,
                            -1./params['Kprime_0']) - 1.)
                 / (params['Kprime_0'] - 1.))
        dintVdpdT = dVTdT*g + VT*dgdKT*params['dKdT_0']
        return (params['S_0'] + self.__intCpoverTdT(temperature, params)
                - dintVdpdT)

    def enthalpy(self, pressure, temperature, volume, params):
        """
        Returns the enthalpy [J/mol] as a function of pressure [Pa]
        and temperature [K].
        """
        gibbs = self.gibbs_free_energy(pressure, temperature, volume, params)
        entropy = self.entropy(pressure, temperature, volume, params)
        return gibbs + temperature * entropy

    def molar_heat_capacity_p(self, pressure, temperature, volume, params):
        """
        Returns the heat capacity [J/K/mol] as a function of pressure [Pa]
        and temperature [K].
        """
        # The differentiation is tedious, so for now we just take the
        # numerical derivative of S
        # TODO calculate the analytical derivative
        dT = 0.1
        S1 = self.entropy(pressure, temperature+dT/2., volume, params)
        S0 = self.entropy(pressure, temperature-dT/2., volume, params)
        return temperature * (S1 - S0)/dT

    def __intCpdT(self, temperature, params):
        """
        Returns the thermal addition to the standard state enthalpy [J/mol]
        at ambient pressure [Pa]
        """
        return ((params['Cp'][0] * temperature
                 + 0.5 * params['Cp'][1] * np.power(temperature, 2.)
                 - params['Cp'][2] / temperature
                 + 2. * params['Cp'][3] * np.sqrt(temperature))
                - (params['Cp'][0] * params['T_0']
                   + 0.5 * params['Cp'][1] * params['T_0'] * params['T_0']
                   - params['Cp'][2] / params['T_0']
                   + 2.0 * params['Cp'][3] * np.sqrt(params['T_0'])))

    def __intCpoverTdT(self, temperature, params):
        """
        Returns the thermal addition to the standard state entropy [J/K/mol]
        at ambient pressure [Pa]
        """
        return ((params['Cp'][0] * np.log(temperature)
                 + params['Cp'][1] * temperature
                 - 0.5 * params['Cp'][2] / np.power(temperature, 2.)
                 - 2.0 * params['Cp'][3] / np.sqrt(temperature))
                - (params['Cp'][0] * np.log(params['T_0'])
                   + params['Cp'][1] * params['T_0']
                   - 0.5 * params['Cp'][2] / (params['T_0'] * params['T_0'])
                   - 2.0 * params['Cp'][3] / np.sqrt(params['T_0'])))

    def validate_parameters(self, params):
        """
        Check for existence and validity of the parameters
        """
        if 'T_0' not in params:
            params['T_0'] = 298.15

        # If standard state enthalpy and entropy are not included
        # this is presumably deliberate, as we can model density
        # and bulk modulus just fine without them.
        # Just add them to the dictionary as nans.
        if 'H_0' not in params:
            params['H_0'] = float('nan')
        if 'S_0' not in params:
            params['S_0'] = float('nan')

        # First, let's check the EoS parameters for Tref
        murn.Murnaghan.validate_parameters(murn.Murnaghan(), params)

        # Now check all the required keys for the
        # thermal part of the EoS are in the dictionary
        expected_keys = ['H_0', 'S_0', 'V_0', 'Cp', 'a_0', 'dKdT_0',
                         'n', 'molar_mass']
        for k in expected_keys:
            if k not in params:
                raise KeyError('params object missing parameter : ' + k)

        # Finally, check that the values are reasonable.
        if params['T_0'] < 0.:
            warnings.warn('Unusual value for T_0', stacklevel=2)
        if ((params['G_0'] is not float('nan')
             and (params['G_0'] < 0. or params['G_0'] > 1.e13))):
            warnings.warn('Unusual value for G_0', stacklevel=2)
        if ((params['Gprime_0'] is not float('nan')
             and (params['Gprime_0'] < -5. or params['Gprime_0'] > 10.))):
            warnings.warn('Unusual value for Gprime_0', stacklevel=2)

        # no test for H_0 or S_0 (several HP endmembers have S_0 < 0)
        if params['V_0'] < 1.e-7 or params['V_0'] > 1.e-2:
            warnings.warn('Unusual value for V_0', stacklevel=2)

        if self.molar_heat_capacity_p0(params['T_0'], params) < 0.:
            warnings.warn('Negative heat capacity at T_0', stacklevel=2)
        if self.molar_heat_capacity_p0(2000., params) < 0.:
            warnings.warn('Negative heat capacity at 2000K', stacklevel=2)

        if params['a_0'] < 0. or params['a_0'] > 1.e-3:
            warnings.warn('Unusual value for a_0', stacklevel=2)

        if params['n'] < 1. or params['n'] > 1000.:
            warnings.warn('Unusual value for n', stacklevel=2)
        if params['molar_mass'] < 0.001 or params['molar_mass'] > 10.:
            warnings.warn('Unusual value for molar_mass', stacklevel=2)



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for the Earth and Planetary Sciences
# Copyright (C) 2012 - 2017 by the BurnMan team, released under the GNU
# GPL v2 or later.

from __future__ import absolute_import
import numpy as np
import scipy.optimize as opt
from ..constants import gas_constant

"""
Functions for modifying the thermodynamic properties of minerals
Currently includes modifications for:
- second order transitions (landau, landau_hp),
- order-disorder (bragg_williams),
- magnetism (magnetic_chs),
- and a linear modification (linear).
"""


def _landau_excesses(pressure, temperature, params):
    """
    Applies a tricritical Landau correction to the properties
    of an endmember which undergoes a displacive phase transition.
    This correction follows Putnis (1992), and is done relative to
    the completely *ordered* state (at 0 K).
    It therefore differs in implementation from both
    Stixrude and Lithgow-Bertelloni (2011) and
    Holland and Powell (2011), who compute properties relative to
    the completely disordered state and standard states respectively.

    The current implementation is preferred, as the excess
    entropy (and heat capacity) terms are equal to zero at 0 K.

    N.B. The excesses are for a *completely relaxed* mineral;
    for example, seismic wave propagation is *slow* compared to the
    rate of reaction.
    """

    Tc = params['Tc_0'] + params['V_D'] * pressure / params['S_D']

    G_disordered = -params['S_D'] * ((temperature - Tc) + params['Tc_0'] / 3.)
    dGdT_disordered = -params['S_D']
    dGdP_disordered = params['V_D']

    if temperature < Tc:
        # Wolfram input to check partial differentials
        # x = T, y = P, a = S, c = Tc0, d = V
        # D[D[a ((x - c - d*y/a)*(1 - x/(c + d*y/a))^0.5 + c/3*(1 - x/(c +
        # d*y/a))^1.5), x], x]
        Q2 = np.sqrt(1. - temperature / Tc)
        G = (params['S_D']
             * ((temperature - Tc) * Q2
                + params['Tc_0'] * Q2 * Q2 * Q2 / 3.) + G_disordered)
        dGdP = (-params['V_D'] * Q2 * (1. + 0.5 * temperature / Tc
                                       * (1. - params['Tc_0'] / Tc))
                + dGdP_disordered)
        dGdT = (params['S_D'] * Q2 * (1.5 - 0.5 * params['Tc_0'] / Tc)
                + dGdT_disordered)
        d2GdP2 = (params['V_D'] * params['V_D'] * temperature
                  / (params['S_D'] * Tc * Tc * Q2)
                  * (temperature * (1. + params['Tc_0'] / Tc) / (4. * Tc)
                     + Q2 * Q2 * (1. - params['Tc_0'] / Tc) - 1.))
        d2GdT2 = -params['S_D'] / (Tc * Q2) * (0.75 - 0.25 * params['Tc_0']
                                               / Tc)
        d2GdPdT = (params['V_D'] / (2. * Tc * Q2)
                   * (1. + (temperature / (2. * Tc) - Q2 * Q2)
                      * (1. - params['Tc_0'] / Tc)))

    else:
        Q2 = 0.
        G = G_disordered
        dGdT = dGdT_disordered
        dGdP = dGdP_disordered
        d2GdT2 = 0.
        d2GdP2 = 0.
        d2GdPdT = 0.

    excesses = {'G': G, 'dGdT': dGdT, 'dGdP': dGdP,
                'd2GdT2': d2GdT2, 'd2GdP2': d2GdP2, 'd2GdPdT': d2GdPdT}

    return (excesses, {'Q': np.sqrt(Q2)})


def _landau_hp_excesses(pressure, temperature, params):
    """
    Applies a tricritical Landau correction to the properties
    of an endmember which undergoes a displacive phase transition.
    This correction is done relative to the standard state, as per
    Holland and Powell (1998).

    Includes the correction published within landaunote.pdf
    (Holland, pers. comm), which 'corrects' the terms involving
    the critical temperature Tc / Tc*

    Note that this formalism is still inconsistent, as it predicts that
    the order parameter can be greater than one. For this reason
    _landau_excesses is preferred.

    N.B. The excesses are for a *completely relaxed* mineral;
    i.e. the seismic wave propagation is *slow* compared to the
    rate of reaction.
    """

    P = pressure
    T = temperature

    if params['T_0'] < params['Tc_0']:
        Q_0 = np.power((params['Tc_0'] - params['T_0']) / params['Tc_0'], 0.25)
    else:
        Q_0 = 0.

    Tc = params['Tc_0'] + params['V_D'] * (P - params['P_0']) / params['S_D']
    if T < Tc:
        Q = np.power((Tc - T) / params['Tc_0'], 0.25)
    else:
        Q = 0.

    # Gibbs
    G = (params['Tc_0'] * params['S_D'] * (Q_0 * Q_0 - np.power(Q_0, 6.) / 3.)
         - params['S_D'] * (Tc * Q * Q - params['Tc_0'] * np.power(Q, 6.) / 3.)
         - T * params['S_D'] * (Q_0 * Q_0 - Q * Q)
         + (P - params['P_0']) * params['V_D'] * Q_0 * Q_0)

    dGdT = params['S_D'] * (Q * Q - Q_0 * Q_0)
    dGdP = -params['V_D'] * (Q * Q - Q_0 * Q_0)

    if Q > 1.e-12:
        d2GdT2 = -params['S_D'] / (2. * params['Tc_0'] * Q * Q)
        d2GdP2 = (-params['V_D'] * params['V_D']
                  / (2. * params['S_D'] * params['Tc_0'] * Q * Q))
        d2GdPdT = params['V_D'] / (2. * params['Tc_0'] * Q * Q)
    else:
        d2GdT2 = 0.
        d2GdP2 = 0.
        d2GdPdT = 0.

    excesses = {'G': G, 'dGdT': dGdT, 'dGdP': dGdP,
                'd2GdT2': d2GdT2, 'd2GdP2': d2GdP2, 'd2GdPdT': d2GdPdT}

    return (excesses, {'Q': Q})


def _linear_excesses(pressure, temperature, params):
    """
    Applies a 'Darken's quadratic formalism' correction (Powell, 1987)
    to the thermodynamic properties of a mineral endmember.
    This correction is relative to P = 0 and T = 0 and linear in P and T
    and therefore corresponds to a constant volume and entropy correction.

    Applying either a volume or entropy term will generally break
    equations of state (i.e. the properties of the mineral will
    no longer obey the equation of state defined in the
    params dictionary. However, this form of excess is extremely
    useful as a first order tweak to free energies
    (especially in solid solution calculations)
    """

    G = params['delta_E'] \
        - (temperature) * params['delta_S'] \
        + (pressure) * params['delta_V']
    dGdT = -params['delta_S']
    dGdP = params['delta_V']
    d2GdT2 = 0.
    d2GdP2 = 0.
    d2GdPdT = 0.

    excesses = {'G': G, 'dGdT': dGdT, 'dGdP': dGdP,
                'd2GdT2': d2GdT2, 'd2GdP2': d2GdP2, 'd2GdPdT': d2GdPdT}

    return (excesses, None)


def _bragg_williams_excesses(pressure, temperature, params):
    """
    Applies a Bragg-Williams type correction to the thermodynamic
    properties of a mineral endmember. Used for modelling
    order-disorder processes.
    Expressions are from Holland and Powell (1996).

    N.B. The excesses are for a *completely relaxed* mineral;
    i.e. the seismic wave propagation is *slow* compared to the
    rate of reaction.

    This may not be reasonable for order-disorder, especially
    for slow or coupled diffusers (Si-Al, for example).
    The completely *unrelaxed* mineral (in terms of order-disorder)
    can be calculated with a solid solution model.
    """

    R = gas_constant
    n = params['n']
    if params['factor'] > 0.:
        f = [params['factor'], params['factor']]
    else:
        f = [1., -params['factor']]

    # Equation A2-2
    def flnarxn(n, Q, f):
        return (n / (n + 1.) * (f[0] * np.log(n*(1. - Q))
                                + f[1] * np.log(1. - Q)
                                - f[0] * np.log(1. + n*Q)
                                - f[1] * np.log(n + Q)))

    # Equation A2-4
    # Can be derived from A2-2 noting that
    # delta_H + f*R*T*lnarxn = delta_G + f*R*T*(lnadisord - lnaord)
    def reaction_bragg_williams(Q, delta_H, temperature, n, f, W):
        return (delta_H + R * temperature * flnarxn(n, Q, f)
                + (2. * Q - 1.) * W)

    def order_gibbs(pressure, temperature, params):
        W = params['Wh'] + pressure * params['Wv']
        H_disord = (params['deltaH']
                    + pressure * params['deltaV'])

        # We can use brentq, but don't let the lower bracket = 0
        try:
            Q = opt.brentq(reaction_bragg_williams, 1.e-12, 1. - 1.e-12,
                           args=(H_disord, temperature, n, f, W))
        except ValueError:
            Q = 0.

        S = - R * (f[0] * ((1. + n*Q)*np.log((1. + n * Q)/(n + 1.))
                    + n * (1. - Q) * np.log(n * (1. - Q) / (n + 1.)))
                   + f[1] * (n * (1. - Q) * np.log((1. - Q) / (n + 1.))
                             + n * (n + Q) * np.log((n + Q) / (n + 1.)))
                   ) / (n + 1.)

        G = (1. - Q)*H_disord + (1. - Q)*Q*W - temperature*S
        return Q, G

    # Calculating partial differentials with respect to P and T
    # are complicated by the fact that Q changes with P and T
    # Since there's no analytical solution for Q(P, T), we are
    # unfortunately driven to numerical differentiation. Schade.
    dT = 0.1
    dP = 1000.

    Q, G = order_gibbs(pressure, temperature, params)
    Q, GsubPsubT = order_gibbs(pressure - dP, temperature - dT, params)
    Q, GsubPaddT = order_gibbs(pressure - dP, temperature + dT, params)
    Q, GaddPsubT = order_gibbs(pressure + dP, temperature - dT, params)
    Q, GaddPaddT = order_gibbs(pressure + dP, temperature + dT, params)
    Q, GsubP = order_gibbs(pressure - dP, temperature, params)
    Q, GaddP = order_gibbs(pressure + dP, temperature, params)
    Q, GsubT = order_gibbs(pressure, temperature - dT, params)
    Q, GaddT = order_gibbs(pressure, temperature + dT, params)

    dGdT = (GaddT - GsubT) / (2. * dT)
    dGdP = (GaddP - GsubP) / (2. * dP)
    d2GdT2 = (GaddT + GsubT - 2. * G) / (dT * dT)
    d2GdP2 = (GaddP + GsubP - 2. * G) / (dP * dP)
    d2GdPdT = (GaddPaddT - GsubPaddT - GaddPsubT + GsubPsubT) / (4. * dT * dP)

    excesses = {'G': G, 'dGdT': dGdT, 'dGdP': dGdP,
                'd2GdT2': d2GdT2, 'd2GdP2': d2GdP2, 'd2GdPdT': d2GdPdT}

    return (excesses, {'Q': Q})


def _magnetic_excesses_chs(pressure, temperature, params):
    """
    Applies a magnetic contribution to the thermodynamic
    properties of a mineral endmember.
    The expression for the gibbs energy contribution is that
    used by Chin, Hertzman and Sundman (1987) as reported
    in the Journal of Phase Equilibria (Sundman, 1991).
    """

    structural_parameter = params['structural_parameter']
    curie_temperature = params['curie_temperature'][
        0] + pressure * params['curie_temperature'][1]
    tau = temperature / curie_temperature
    dtaudT = 1. / curie_temperature
    dtaudP = -(temperature * params['curie_temperature'][1]) / (
        curie_temperature * curie_temperature)
    d2taudPdT = params['curie_temperature'][
        1] / (curie_temperature * curie_temperature)
    d2taudP2 = (2. * temperature * params['curie_temperature'][1]
                * params['curie_temperature'][1]
                / (curie_temperature * curie_temperature * curie_temperature))
    magnetic_moment = params['magnetic_moment'][
        0] + pressure * params['magnetic_moment'][1]
    dmagnetic_momentdP = params['magnetic_moment'][1]

    A = (518. / 1125.) + (11692. / 15975.) * ((1. / structural_parameter) - 1.)
    if tau < 1:
        f = 1. - (1. / A) * (79. / (140. * structural_parameter * tau)
                             + (474. / 497.) * (1. / structural_parameter - 1.)
                             * (np.power(tau, 3.) / 6.
                                + np.power(tau, 9.) / 135.
                                + np.power(tau, 15.) / 600.))
        dfdtau = -(1. / A) * (-79. / (140. * structural_parameter * tau * tau)
                              + (474. / 497.) * (1. / structural_parameter
                                                 - 1.)
                              * (tau * tau / 2.
                                 + np.power(tau, 8.) / 15.
                                 + np.power(tau, 14.) / 40.))
        d2fdtau2 = -(1. / A) * (2. * 79. / (140. * structural_parameter
                                            * np.power(tau, 3.))
                                + (474. / 497.) * (1. / structural_parameter
                                                   - 1.)
                                * (tau
                                   + 8. * np.power(tau, 7.) / 15.
                                   + 14. * np.power(tau, 13.) / 40.))

    else:
        f = - (1. / A) * (np.power(tau, -5.) / 10. + np.power(
                tau, -15.) / 315. + np.power(tau, -25.) / 1500.)
        dfdtau = (1. / A) * (np.power(tau, -6.) / 2.
                             + np.power(tau, -16.) / 21.
                             + np.power(tau, -26.)
                             / 60.)
        d2fdtau2 = - (1. / A) * (6. * np.power(tau, -7.) / 2. + 16.
                                 * np.power(tau, -17.) / 21. + 26.
                                 * np.power(tau, -27.)
                                 / 60.)

    dfdT = dfdtau * dtaudT
    d2fdT2 = d2fdtau2 * dtaudT * dtaudT
    dfdP = dfdtau * dtaudP
    d2fdP2 = d2fdtau2 * dtaudP * dtaudP + dfdtau * d2taudP2
    d2fdPdT = d2fdtau2 * dtaudT * dtaudP - dfdtau * d2taudPdT

    G = gas_constant * temperature * np.log(magnetic_moment + 1.) * f
    dGdT = gas_constant * \
        np.log(magnetic_moment + 1.) * (f + temperature * dfdT)
    d2GdT2 = gas_constant * \
        np.log(magnetic_moment + 1.) * (2. * dfdT + temperature * d2fdT2)

    dGdP = gas_constant * temperature * (f * dmagnetic_momentdP
                                         / (magnetic_moment + 1.)
                                         + dfdP * np.log(magnetic_moment + 1.))
    d2GdP2 = gas_constant * temperature * (-f * np.power(dmagnetic_momentdP
                                                         / (magnetic_moment + 1.), 2.)
                                           + 2 * dfdP * dmagnetic_momentdP
                                           / (magnetic_moment + 1.)
                                           + d2fdP2 * np.log(magnetic_moment + 1.))
    d2GdPdT = dGdP / temperature + (gas_constant * temperature
                                    * np.log(magnetic_moment + 1.)
                                    * d2fdPdT
                                    + gas_constant * temperature
                                    * dmagnetic_momentdP
                                    / (magnetic_moment + 1.) * dfdT)

    excesses = {'G': G, 'dGdT': dGdT, 'dGdP': dGdP,
                'd2GdT2': d2GdT2, 'd2GdP2': d2GdP2, 'd2GdPdT': d2GdPdT}

    return (excesses, None)


def calculate_property_modifications(mineral):
    """
    Sums the excesses from all the modifiers.

    To calculate thermodynamic properties from the outputs,
    the following functions should be used
    (the _o suffix stands for original value):

    gibbs = gibbs_o + excesses['G']
    S = S_o - excesses['dGdT']
    V = V_o + excesses['dGdP']
    K_T = V / ((V_o / K_T_o) - excesses['d2GdP2'])
    C_p = C_p_o - temperature*excesses['d2GdT2']
    alpha = ((alpha_o*V_o) + excesses['d2GdPdT']) / V

    H = gibbs + temperature*S
    helmholtz = gibbs - pressure*V
    C_v = C_p - V*temperature*alpha*alpha*K_T
    gr = alpha*K_T*V/C_v
    K_S = K_T*C_p/C_v
    """
    excesses = {'G': 0., 'dGdT': 0., 'dGdP': 0.,
                'd2GdT2': 0., 'd2GdP2': 0., 'd2GdPdT': 0.}
    mineral.property_modifier_properties = []
    for modifier in mineral.property_modifiers:
        if modifier[0] == 'landau':
            xs_function = _landau_excesses
        if modifier[0] == 'landau_hp':
            xs_function = _landau_hp_excesses
        if modifier[0] == 'linear':
            xs_function = _linear_excesses
        if modifier[0] == 'bragg_williams':
            xs_function = _bragg_williams_excesses
        if modifier[0] == 'magnetic_chs':
            xs_function = _magnetic_excesses_chs

        xs_component, properties = xs_function(
            mineral.pressure, mineral.temperature, modifier[1])
        mineral.property_modifier_properties.append(properties)
        for key in xs_component:
            excesses[key] += xs_component[key]

    return excesses



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for the Earth and Planetary Sciences
# Copyright (C) 2012 - 2017 by the BurnMan team, released under the GNU
# GPL v2 or later.

from __future__ import absolute_import

import numpy as np
import scipy.optimize as opt
import warnings

from . import equation_of_state as eos
from . import birch_murnaghan as bm
from . import debye
from .. import constants
from ..tools.math import bracket


class MGDBase(eos.EquationOfState):

    """
    Base class for a generic finite-strain Mie-Grueneisen-Debye
    equation of state.  References for this can be found in many
    places, such as Shim, Duffy and Kenichi (2002) and Jackson and Rigden
    (1996).  Here we mostly follow the appendices of Matas et al (2007).
    Of particular note is the thermal correction to the shear modulus, which
    was developed by Hama and Suito (1998).
    """

    def grueneisen_parameter(self, pressure, temperature, volume, params):
        """
        Returns grueneisen parameter [unitless] as a function of pressure,
        temperature, and volume (EQ B6)
        """
        return self._grueneisen_parameter(params['V_0'] / volume, params)

    def volume(self, pressure, temperature, params):
        """
        Returns volume [m^3] as a function of pressure [Pa] and temperature [K]
        EQ B7
        """
        T_0 = params['T_0']
        func = lambda x: bm.birch_murnaghan(params['V_0'] / x, params) + \
            self._thermal_pressure(temperature, x, params) - \
            self._thermal_pressure(T_0, x, params) - pressure
        try:
            sol = bracket(func, params['V_0'], 1.e-2 * params['V_0'])
        except:
            raise ValueError(
                'Cannot find a volume, perhaps you are outside of the range of validity for the equation of state?')
        return opt.brentq(func, sol[0], sol[1])

    def isothermal_bulk_modulus(self, pressure, temperature, volume, params):
        """
        Returns isothermal bulk modulus [Pa] as a function of pressure [Pa],
        temperature [K], and volume [m^3].  EQ B8
        """
        T_0 = params['T_0']
        K_T = bm.bulk_modulus(volume, params) + \
            self._thermal_bulk_modulus(temperature, volume, params) - \
            self._thermal_bulk_modulus(T_0, volume, params)  # EQB13
        return K_T

    # calculate the mgd shear modulus as a function of P, V, and T
    def shear_modulus(self, pressure, temperature, volume, params):
        """
        Returns shear modulus [Pa] as a function of pressure [Pa],
        temperature [K], and volume [m^3].  EQ B11
        """
        T_0 = params['T_0']
        if self.order == 2:
            return bm.shear_modulus_second_order(volume, params) + \
                self._thermal_shear_modulus(temperature, volume, params) - \
                self._thermal_shear_modulus(T_0, volume, params)  # EQ B11
        elif self.order == 3:
            return bm.shear_modulus_third_order(volume, params) + \
                self._thermal_shear_modulus(temperature, volume, params) - \
                self._thermal_shear_modulus(T_0, volume, params)  # EQ B11
        else:
            raise NotImplementedError("")

    # heat capacity at constant volume
    def molar_heat_capacity_v(self, pressure, temperature, volume, params):
        """
        Returns heat capacity at constant volume at the pressure, temperature, and volume [J/K/mol]
        """
        Debye_T = self._debye_temperature(params['V_0'] / volume, params)
        C_v = debye.molar_heat_capacity_v(temperature, Debye_T, params['n'])
        return C_v

    def thermal_expansivity(self, pressure, temperature, volume, params):
        """
        Returns thermal expansivity at the pressure, temperature, and volume [1/K]
        """
        C_v = self.molar_heat_capacity_v(pressure, temperature, volume, params)
        gr = self._grueneisen_parameter(params['V_0'] / volume, params)
        K = self.isothermal_bulk_modulus(pressure, temperature, volume, params)
        alpha = gr * C_v / K / volume
        return alpha

    # heat capacity at constant pressure
    def molar_heat_capacity_p(self, pressure, temperature, volume, params):
        """
        Returns heat capacity at constant pressure at the pressure, temperature, and volume [J/K/mol]
        """
        alpha = self.thermal_expansivity(pressure, temperature, volume, params)
        gr = self._grueneisen_parameter(params['V_0'] / volume, params)
        C_v = self.molar_heat_capacity_v(pressure, temperature, volume, params)
        C_p = C_v * (1. + gr * alpha * temperature)
        return C_p

    def adiabatic_bulk_modulus(self, pressure, temperature, volume, params):
        """
        Returns adiabatic bulk modulus [Pa] as a function of pressure [Pa],
        temperature [K], and volume [m^3].  EQ D6
        """
        K_T = self.isothermal_bulk_modulus(
            pressure, temperature, volume, params)
        alpha = self.thermal_expansivity(pressure, temperature, volume, params)
        gr = self._grueneisen_parameter(params['V_0'] / volume, params)
        K_S = K_T * (1. + gr * alpha * temperature)
        return K_S

    def pressure(self, temperature, volume, params):
        """
        Returns pressure [Pa] as a function of temperature [K] and volume[m^3]
        EQ B7
        """
        T_0 = params['T_0']
        return bm.birch_murnaghan(params['V_0'] / volume, params) + \
            self._thermal_pressure(temperature, volume, params) - \
            self._thermal_pressure(T_0, volume, params)

    def gibbs_free_energy(self, pressure, temperature, volume, params):
        """
        Returns the Gibbs free energy at the pressure and temperature of the mineral [J/mol]
        """
        G = self.helmholtz_free_energy(
            pressure, temperature, volume, params) + pressure * volume
        return G

    def molar_internal_energy(self, pressure, temperature, volume, params):
        """
        Returns the internal energy at the pressure and temperature of the mineral [J/mol]
        """
        return self.helmholtz_free_energy(pressure, temperature, volume, params) + \
            temperature * \
            self.entropy(pressure, temperature, volume, params)

    def entropy(self, pressure, temperature, volume, params):
        """
        Returns the entropy at the pressure and temperature of the mineral [J/K/mol]
        """
        Debye_T = self._debye_temperature(params['V_0'] / volume, params)
        S = debye.entropy(temperature, Debye_T, params['n'])
        return S

    def enthalpy(self, pressure, temperature, volume, params):
        """
        Returns the enthalpy at the pressure and temperature of the mineral [J/mol]
        """

        return self.helmholtz_free_energy(pressure, temperature, volume, params) + \
            temperature * self.entropy(pressure, temperature, volume, params) + \
            pressure * volume

    def helmholtz_free_energy(self, pressure, temperature, volume, params):
        """
        Returns the Helmholtz free energy at the pressure and temperature of the mineral [J/mol]
        """
        x = params['V_0'] / volume
        f = 1. / 2. * (pow(x, 2. / 3.) - 1.)
        b_iikk = 9. * params['K_0']  # EQ 28, SLB2005
        b_iikkmm = 27. * params['K_0'] * (params['Kprime_0'] - 4.)  # EQ 29, SLB2005

        F_pressure = ( 0.5 * b_iikk * f * f * params['V_0'] +
                       (1. / 6.) * params['V_0'] * b_iikkmm * f * f * f )

        Debye_T = self._debye_temperature(params['V_0'] / volume, params)
        F_thermal = debye.helmholtz_free_energy(temperature, Debye_T, params['n']) - \
                    debye.helmholtz_free_energy(params['T_0'], Debye_T, params['n'])

        return params['F_0'] + F_pressure + F_thermal

    # calculate the thermal correction to the shear modulus as a function of
    # V, T
    def _thermal_shear_modulus(self, T, V, params):
        if T > 1.e-10:
            gr = self._grueneisen_parameter(params['V_0'] / V, params)
            Debye_T = self._debye_temperature(params['V_0'] / V, params)
            G_th = 3. / 5. * (self._thermal_bulk_modulus(T, V, params) -
                              6 * constants.gas_constant * T * params['n'] / V * gr * debye.debye_fn(Debye_T / T))  # EQ B10
            return G_th
        else:
            return 0.

    # compute the Debye temperature in K.  Takes the
    # parameter x, which is V_0/V (molar volumes).
    # Depends on the reference grueneisen parameter,
    # the reference Debye temperature, and the factor
    # q_0, see Matas eq B6
    def _debye_temperature(self, x, params):
        return params['Debye_0'] * np.exp((params['grueneisen_0'] -
                                           self._grueneisen_parameter(x, params)) / params['q_0'])

    # compute the grueneisen parameter with depth, according
    # to q_0.  Takes x=V_0/V. See Matas eq B6
    def _grueneisen_parameter(self, x, params):
        return params['grueneisen_0'] * pow(1. / x, params['q_0'])

    # calculate isotropic thermal pressure, see
    # Matas et. al. (2007) eq B4
    def _thermal_pressure(self, T, V, params):
        Debye_T = self._debye_temperature(params['V_0'] / V, params)
        gr = self._grueneisen_parameter(params['V_0'] / V, params)
        P_th = gr * debye.thermal_energy(T, Debye_T, params['n']) / V
        return P_th

    # calculate the thermal correction for the mgd
    # bulk modulus (see matas et al, 2007)
    def _thermal_bulk_modulus(self, T, V, params):
        if T > 1.e-10:
            gr = self._grueneisen_parameter(params['V_0'] / V, params)
            Debye_T = self._debye_temperature(params['V_0'] / V, params)
            K_th = 3. * params['n'] * constants.gas_constant * T / V * gr * \
                ((1. - params['q_0'] - 3. * gr) * debye.debye_fn(
                 Debye_T / T) + 3. * gr * (Debye_T / T) / (np.exp(Debye_T / T) - 1.))  # EQ B5
            return K_th
        else:
            return 0.

    def validate_parameters(self, params):
        """
        Check for existence and validity of the parameters
        """
        if 'T_0' not in params:
            params['T_0'] = 300.
        if 'F_0' not in params:
            params['F_0'] = 0.

        # First, let's check the EoS parameters for Tref
        bm.BirchMurnaghanBase.validate_parameters(
            bm.BirchMurnaghanBase(), params)

        # Now check all the required keys for the
        # thermal part of the EoS are in the dictionary
        expected_keys = ['molar_mass', 'n', 'Debye_0', 'grueneisen_0', 'q_0']
        for k in expected_keys:
            if k not in params:
                raise KeyError('params object missing parameter : ' + k)

        # Finally, check that the values are reasonable.
        if params['T_0'] < 0.:
            warnings.warn('Unusual value for T_0', stacklevel=2)
        if params['molar_mass'] < 0.001 or params['molar_mass'] > 1.:
            warnings.warn('Unusual value for molar_mass', stacklevel=2)
        if params['n'] < 1. or params['n'] > 1000.:
            warnings.warn('Unusual value for n', stacklevel=2)
        if params['Debye_0'] < 1. or params['Debye_0'] > 10000.:
            warnings.warn('Unusual value for Debye_0', stacklevel=2)
        if params['grueneisen_0'] < 0. or params['grueneisen_0'] > 10.:
            warnings.warn('Unusual value for grueneisen_0', stacklevel=2)
        if params['q_0'] < -10. or params['q_0'] > 10.:
            warnings.warn('Unusual value for q_0', stacklevel=2)


class MGD3(MGDBase):

    """
    MGD equation of state with third order finite strain expansion for the
    shear modulus (this should be preferred, as it is more thermodynamically
    consistent.
    """

    def __init__(self):
        self.order = 3


class MGD2(MGDBase):

    """
    MGD equation of state with second order finite strain expansion for the
    shear modulus.  In general, this should not be used, but sometimes
    shear modulus data is fit to a second order equation of state.  In that
    case, you should use this.  The moral is, be careful!
    """

    def __init__(self):
        self.order = 2



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for
# the Earth and Planetary Sciences
# Copyright (C) 2012 - 2019 by the BurnMan team, released under the GNU
# GPL v2 or later.


from __future__ import absolute_import
import numpy as np
from .. import constants


def calculate_nakajima_fp_pv_partition_coefficient(pressure, temperature,
                                                   bulk_composition_mol, initial_distribution_coefficient):
    """
    Calculate the partitioning of iron between periclase and bridgmanite as given
    by Nakajima et al., 2012.

    Parameters
    ----------
    pressure : float
        Equilibrium pressure [Pa]
    temperature : float
        Equilibrium temperature [K]
    bulk_composition_mol : dictionary
        Bulk composition [mol].
        Only Mg, Fe, and Si are assumed to explicitly affect the partitioning
        with Al playing an implicit role.
    initial_distribution_coefficient : float
        The distribution coefficient (Kd_0) at 25 GPa and 0 K

    Returns
    -------
    partition_coefficient_array : tuple
        The proportion of Fe in ferropericlase and perovskite, respectively
    """

    norm = bulk_composition_mol['Mg'] + bulk_composition_mol['Fe']
    f_FeO = bulk_composition_mol['Fe']/norm
    f_SiO2 = bulk_composition_mol['Si']/norm

    Kd_0 = initial_distribution_coefficient
    delV = 2.e-7  # in m^3/mol, average taken from Nakajima et al 2012, JGR

    # eq 5 Nakajima et al 2012, JGR. Solved for ln(K(P,T,X))
    rs = ((25.e9 - pressure) * delV
          / (constants.gas_constant * temperature)) + np.log(Kd_0)

    # The exchange coefficent at P and T. K(P,T,X) in eq 5 Nakajima et al 2012
    K = np.exp(rs)

    # Solving equation 6 in Nakajima et al., 2012 for X_Fe_fp and X_Fe_pv
    # Solved using the definition of the distribution coefficient to define X_Fe_fp as a function of X_Fe_pv

    num_to_sqrt = ((-4. * f_FeO * (K - 1.) * K * f_SiO2)
                   + (pow(1. + (f_FeO * (K - 1)) + ((K - 1.) * f_SiO2), 2.)))

    X_Fe_pv = ((-1. + f_FeO - (f_FeO * K) + f_SiO2 - (f_SiO2 * K) + np.sqrt(num_to_sqrt))
               / (2. * f_SiO2 * (1. - K)))

    X_Fe_fp = X_Fe_pv / (((1. - X_Fe_pv) * K) + X_Fe_pv)

    return (X_Fe_fp, X_Fe_pv)



from __future__ import absolute_import
import re

versionstuff = dict(
    re.findall("(.+) = '(.+)'\n", open('burnman/version.py').read()))

metadata = dict(name='burnman',
                version=versionstuff['version'],
                description='A thermoelastic and thermodynamic toolkit for Earth and planetary sciences',
                url='https://geodynamics.github.io/burnman/',
                author='The BurnMan Team',
                author_email='bob.myhill@bristol.ac.uk',
                license='GPL',
                long_description_content_type="text/x-rst",
                long_description='BurnMan is a Python library for generating thermodynamic and thermoelastic models of planetary interiors.',
                packages=['burnman', 'burnman.minerals', 'burnman.eos', 'burnman.tools', 'burnman.classes'],
                package_data={'burnman': ['data/input_*/*']},
                classifiers=[
                'License :: OSI Approved :: GNU General Public License v2 or later (GPLv2+)',
                'Programming Language :: Python :: 3.7',
                'Programming Language :: Python :: 3.8',
                'Programming Language :: Python :: 3.8'],
                )

# Try to use setuptools in order to check dependencies.
# if the system does not have setuptools, fall back on
# distutils.
try:
    from setuptools import setup
    metadata['install_requires'] = ['numpy', 'matplotlib', 'scipy', 'sympy', 'cvxpy']
except ImportError:
    from distutils.core import setup


setup(**metadata)



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for the Earth and Planetary Sciences
# Copyright (C) 2012 - 2017 by the BurnMan team, released under the GNU
# GPL v2 or later.

from __future__ import absolute_import

import numpy as np
from .. import constants

"""
Functions for the Einstein model of a solid.
"""

eps = np.finfo(float).eps


def thermal_energy(T, einstein_T, n):
    """
    calculate the thermal energy of a substance.  Takes the temperature,
    the Einstein temperature, and n, the number of atoms per molecule.
    Returns thermal energy in J/mol
    """
    if T <= eps:
        # zero point energy
        return 3. * n * constants.gas_constant * einstein_T * 0.5
    x = einstein_T / T
    E_th = 3. * n * constants.gas_constant * einstein_T * \
        (0.5 + 1. / (np.exp(x) - 1.0))  # include the zero point energy
    return E_th


def molar_heat_capacity_v(T, einstein_T, n):
    """
    Heat capacity at constant volume.  In J/K/mol
    """
    if T <= eps:
        return 0.
    x = einstein_T / T
    C_v = 3.0 * n * constants.gas_constant * \
        (x * x * np.exp(x) / np.power(np.exp(x) - 1.0, 2.0))
    return C_v



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for
# the Earth and Planetary Sciences
# Copyright (C) 2012 - 2021 by the BurnMan team, released under the GNU
# GPL v2 or later.

from __future__ import absolute_import
from __future__ import print_function
import numpy as np

from scipy.linalg import lu_factor, lu_solve
from collections import namedtuple

from ..optimize.nonlinear_solvers import damped_newton_solve
from ..classes.solidsolution import SolidSolution


def calculate_constraints(assemblage):
    """
    This function calculates the linear inequality constraints bounding
    the valid parameter space for a given assemblage.

    The constraints are as follows:

      - Pressure and temperature must be positive
      - All phase fractions must be positive
      - All site-species occupancies must be positive

    The constraints are stored in a vector (b) and matrix (A).
    The sign convention is chosen such that the constraint is satisfied
    if A.x + b < eps.

    Parameters
    ----------
    assemblage : burnman.Composite object
        The assemblage for which the constraints are calculated.

    Returns
    -------
    c_vector :
        The constraints vector.

    c_matrix :
        The constraints matrix.
    """
    bounds = []
    n_constraints = 0
    for i, n in enumerate(assemblage.endmembers_per_phase):
        n_constraints += 1
        if n == 1:
            bounds.append(np.array([[]]))
        else:
            bounds.append(assemblage.phases[i].solution_model.endmember_occupancies)
            n_constraints += len(bounds[-1][0])

    c_vector = np.zeros((n_constraints+2))
    c_matrix = np.zeros((n_constraints+2, assemblage.n_endmembers+2))  # includes P, T

    c_matrix[0, 0] = -1  # P>0
    c_matrix[1, 1] = -1  # T>0

    cidx = 2  # index of current compositional constraint
    pidx = 0  # starting index of current phase
    for i, n in enumerate(assemblage.endmembers_per_phase):
        m = len(bounds[i][0])
        # The first endmember proportion is not a free variable
        # (all endmembers in any solution must sum to one)
        # Re-express the constraints without the first endmember
        c_matrix[cidx, pidx+2] = -1.  # need phase proportions > 0
        cidx += 1
        if m != 0:
            c_vector[cidx:cidx+m] = -bounds[i][0]
            c_matrix[cidx:cidx+m,
                     pidx+1+2:pidx+n+2] = (np.einsum('i, j', bounds[i][0],
                                                     np.ones_like(bounds[i][1:, 0]))
                                           - bounds[i].T[:, 1:])
            cidx += m
        pidx += n

    return c_vector, c_matrix


def get_parameters(assemblage):
    """
    Gets the parameters vector (x) for the current equilibrium problem.
    These are:

      - pressure
      - temperature
      - absolute amount of each phase. if a phase is a solution
        with >1 endmember, the following parameters are the mole fractions
        of the independent endmembers in the solution, except for the first
        endmember (as the mole fractions must sum to one).

    Parameters
    ----------
    assemblage : burnman.Composite object
        The assemblage for which equilibrium is to be calculated.

    Returns
    -------
    params : numpy array
        An array containing the current parameter values.
    """
    params = np.empty(assemblage.n_endmembers + 2)
    n_moles_phase = assemblage.n_moles * np.array(assemblage.molar_fractions)

    try:
        params[:2] = [assemblage.pressure, assemblage.temperature]
    except AttributeError:
        raise Exception('You need to set_state before getting parameters')

    j=2
    for i, ph in enumerate(assemblage.phases):
        params[j] = n_moles_phase[i]
        if isinstance(ph, SolidSolution):
            params[j+1:j+assemblage.endmembers_per_phase[i]] = assemblage.phases[i].molar_fractions[1:]
        j += assemblage.endmembers_per_phase[i]

    return params


def get_endmember_amounts(assemblage):
    """
    Gets the absolute amounts of all the endmembers in the solution.

    Parameters
    ----------
    assemblage : burnman.Composite object
        The assemblage for which equilibrium is to be calculated.

    Returns
    -------
    amounts : numpy array
        An array containing the current amounts of all the endmembers.
    """
    phase_amounts = assemblage.n_moles * assemblage.molar_fractions
    amounts = np.empty(assemblage.n_endmembers)
    j=0
    for i, ph in enumerate(assemblage.phases):
        if isinstance(ph, SolidSolution):
            amounts[j:j+assemblage.endmembers_per_phase[i]] = phase_amounts[i] * assemblage.phases[i].molar_fractions
        else:
            amounts[j] = phase_amounts[i]
        j += assemblage.endmembers_per_phase[i]

    return amounts


def set_compositions_and_state_from_parameters(assemblage, parameters):
    """
    Sets the phase compositions, amounts and state of the assemblage
    from a list of parameter values.

    Parameters
    ----------
    assemblage : burnman.Composite object
        The assemblage for which equilibrium is to be calculated.

    parameters : numpy array
        An array containing the current parameter values.
    """
    assemblage.set_state(parameters[0], parameters[1])
    i = 2
    phase_amounts = np.zeros(len(assemblage.phases))
    for phase_idx, ph in enumerate(assemblage.phases):
        phase_amounts[phase_idx] = parameters[i]
        if isinstance(ph, SolidSolution):
            n_mbrs = len(ph.endmembers)
            f = [0.]*n_mbrs
            f[1:] = parameters[i+1:i+n_mbrs]
            f[0] = 1. - sum(f)
            ph.set_composition(f)
            i += n_mbrs
        else:
            i += 1

    assert(np.all(phase_amounts > -1.e-8))
    phase_amounts = np.abs(phase_amounts)
    assemblage.n_moles = sum(phase_amounts)
    assemblage.set_fractions(phase_amounts/assemblage.n_moles)
    return None


def F(x, assemblage, equality_constraints, reduced_composition_vector):
    """
    Returns a vector of values which are zero at equilibrium.
    The first two vector values depend on the
    equality_constraints chosen. For example, if
      - eq[i][0] = 'P', F[i] = P - eq[i][1]
      - eq[i][0] = 'T', F[i] = T - eq[i][1]
      - eq[i][0] = 'S', F[i] = entropy - eq[i][1]
      - eq[i][0] = 'V', F[i] = volume - eq[i][1]
      - eq[i][0] = 'PT_ellipse', F[i] = norm(([P, T] - eq[i][1][0])/eq[i][1][1]) - 1
      - eq[i][0] = 'X', np.dot(eq[i][1][0], x) - eq[i][1][1]

    The next set of vector values correspond to the reaction affinities.
    The final set of vector values correspond to the bulk
    composition constraints.

    Parameters
    ----------
    x : numpy array
        Parameter values for the equilibrium problem to be solved.

    assemblage : burnman.Composite object
        The assemblage for which equilibrium is to be calculated.

    equality_constraints : list of lists
        A list of the equality constraints (see above).

    reduced_composition_vector : numpy array
        The vector corresponding to the amounts of the independent
        elements.

    Returns
    -------
    eqns : numpy array
        An array containing the vector values which
        are equal to zero at equilibrium.
    """

    set_compositions_and_state_from_parameters(assemblage, x)
    new_endmember_amounts = get_endmember_amounts(assemblage)

    # We want to find the root of the following equations
    eqns = np.zeros((2 + assemblage.n_endmembers))
    for i, (type_c, eq_c) in enumerate(equality_constraints):
        if type_c == 'P':
            eqns[i] = x[0] - eq_c
        elif type_c == 'T':
            eqns[i] = x[1] - eq_c
        elif type_c == 'S':
            eqns[i] = assemblage.molar_entropy*assemblage.n_moles - eq_c
        elif type_c == 'V':
            eqns[i] = assemblage.molar_volume*assemblage.n_moles - eq_c
        elif type_c == 'PT_ellipse':
            v_scaled = (x[0:2] - eq_c[0])/eq_c[1]
            eqns[i] = np.linalg.norm(v_scaled) - 1.
        elif type_c == 'X':
            eqns[i] = np.dot(eq_c[0], x) - eq_c[1]  # i.e. Ax = b
        else:
            raise Exception('constraint type not recognised')

    eqns[2:2+assemblage.n_reactions] = assemblage.reaction_affinities
    eqns[2+assemblage.n_reactions:] = (np.dot(assemblage.reduced_stoichiometric_array.T,
                                              new_endmember_amounts)
                                       - reduced_composition_vector)
    return eqns


def jacobian(x, assemblage, equality_constraints):
    """
    The Jacobian of the equilibrium problem (dF/dx).
    See documentation for F and get_parameters
    (which return F and x respectively) for more details.

    Parameters
    ----------
    x : numpy array
        Parameter values for the equilibrium problem to be solved.

    assemblage : burnman.Composite object
        The assemblage for which equilibrium is to be calculated.

    equality_constraints : list of lists
        A list of the equality constraints (see documentation for F).

    Returns
    -------
    jacobian : 2D numpy array
        An array containing the Jacobian for the equilibrium problem.
    """
    # The solver always calls the Jacobian with the same
    # x parameters as used previously for the root functions
    # Therefore we don't need to set compositions or state again here.

    # First, we find out the effect of the two constraint parameters F[:2]
    # on the pressure (x[0]) and temperature (x[1]):
    jacobian = np.zeros((assemblage.n_endmembers+2, assemblage.n_endmembers+2))
    for i, (type_c, eq_c) in enumerate(equality_constraints):
        if type_c == 'P':  # dP/dx
            jacobian[i, 0] = 1.  # jacobian[i, j!=0] = 0
        elif type_c == 'T':  # dT/dx
            jacobian[i, 1] = 1.  # jacobian[i, j!=1] = 0
        elif type_c == 'S':  # dS/dx
            # dS/dP = -aV, dS/dT = Cp/T
            jacobian[i, 0:2] = [-assemblage.n_moles
                                * assemblage.alpha
                                * assemblage.molar_volume,
                                assemblage.n_moles
                                * assemblage.molar_heat_capacity_p / x[1]]
            j = 2
            for k, n in enumerate(assemblage.endmembers_per_phase):
                jacobian[i, j] = assemblage.phases[k].molar_entropy
                if n > 1:  # for solutions with >1 endmember
                    jacobian[i, j+1:j+n] = (assemblage.n_moles
                                            * assemblage.molar_fractions[k]
                                            * (assemblage.phases[k].partial_entropies[1:]
                                               - assemblage.phases[k].partial_entropies[0]))
                j += n
        elif type_c == 'V':  # dV/dx
            # dV/dP = -V/K_T, dV/dT = aV
            jacobian[i, 0:2] = [-assemblage.n_moles
                                * assemblage.molar_volume / assemblage.K_T,
                                assemblage.n_moles*assemblage.molar_volume]
            j = 2
            for k, n in enumerate(assemblage.endmembers_per_phase):
                jacobian[i, j] = assemblage.phases[k].molar_volume
                if n > 1:  # for solutions with >1 stable endmember
                    jacobian[i, j+1:j+n] = (assemblage.n_moles
                                            * assemblage.molar_fractions[k]
                                            * (assemblage.phases[k].partial_volumes[1:]
                                               - assemblage.phases[k].partial_volumes[0]))
                j += n
        elif type_c == 'PT_ellipse':
            v_scaled = (x[0:2] - eq_c[0])/eq_c[1]
            jacobian[i, 0:2] = v_scaled/(np.linalg.norm(v_scaled)*eq_c[1])
        elif type_c == 'X':
            jacobian[i, :] = eq_c[0]
        else:
            raise Exception('constraint type not recognised')

    # Next, let's get the effect of pressure and temperature
    # on each of the independent reactions
    # i.e. dF(i, reactions)/dx[0] and dF(i, reactions)/dx[1]
    partial_volumes_vector = np.zeros((assemblage.n_endmembers))
    partial_entropies_vector = np.zeros((assemblage.n_endmembers))
    j = 0
    for i, n in enumerate(assemblage.endmembers_per_phase):
        if n == 1:  # for endmembers
            partial_volumes_vector[j] = assemblage.phases[i].molar_volume
            partial_entropies_vector[j] = assemblage.phases[i].molar_entropy
        else:  # for solid solutions
            partial_volumes_vector[j:j+n] = assemblage.phases[i].partial_volumes
            partial_entropies_vector[j:j+n] = assemblage.phases[i].partial_entropies
        j += n
    reaction_volumes = np.dot(assemblage.reaction_basis, partial_volumes_vector)
    reaction_entropies = np.dot(assemblage.reaction_basis, partial_entropies_vector)

    # dGi/dP = deltaVi; dGi/dT = -deltaSi
    jacobian[2:2+len(reaction_volumes), 0] = reaction_volumes
    jacobian[2:2+len(reaction_volumes), 1] = -reaction_entropies

    # Pressure and temperature have no effect on the bulk
    # compositional constraints
    # i.e. dF(i, bulk)/dx[0] and dF(i, bulk)/dx[1] = 0

    # Finally, let's build the compositional Hessian d2G/dfidfj = dmui/dfj
    # where fj is the fraction of endmember j in a phase
    phase_amounts = np.array(assemblage.molar_fractions) * assemblage.n_moles
    comp_hessian = np.zeros((assemblage.n_endmembers, assemblage.n_endmembers))
    dfi_dxj = np.zeros((assemblage.n_endmembers, assemblage.n_endmembers))
    dpi_dxj = np.zeros((assemblage.n_endmembers, assemblage.n_endmembers))
    j = 0
    for i, n in enumerate(assemblage.endmembers_per_phase):
        if n == 1:
            # changing the amount (p) of a pure phase
            # does not change its fraction in that phase,
            # so dfi_dxj remains unchanged
            dpi_dxj[j, j] = 1.
        else:
            comp_hessian[j:j+n, j:j+n] = assemblage.phases[i].gibbs_hessian

            # x[0] = p(phase) and x[1:] = f[1:] - f[0]
            # Therefore
            # df[0]/dx[0] = 0
            # df[0]/dx[1:] = -1
            # (because changing the fraction of any endmember
            # depletes the fraction of the first endmember)
            # df[1:]/dx[1:] = 1 on diagonal, 0 otherwise
            # (because all other fractions are independent of each other)
            dfi_dxj[j:j+n, j:j+n] = np.eye(n)
            dfi_dxj[j, j:j+n] -= 1.
            # Total amounts of endmembers (p) are the fractions
            # multiplied by the amounts of their representative phases
            dpi_dxj[j:j+n, j:j+n] = dfi_dxj[j:j+n, j:j+n] * phase_amounts[i]
            # The derivative of the amount of each endmember with respect
            # to the amount of each phase is equal to the molar fractions
            # of the endmembers.
            dpi_dxj[j:j+n, j] = assemblage.phases[i].molar_fractions
        j += n

    # dfi_dxj converts the endmember hessian to the parameter hessian.
    reaction_hessian = assemblage.reaction_basis.dot(comp_hessian).dot(dfi_dxj)
    bulk_hessian = assemblage.reduced_stoichiometric_array.T.dot(dpi_dxj)
    jacobian[2:, 2:] = np.concatenate((reaction_hessian, bulk_hessian))
    return jacobian


def lambda_bounds(dx, x, endmembers_per_phase):
    """
    Returns the lambda bounds for the damped affine invariant modification
    to Newton's method for nonlinear problems (Deuflhard, 1974;1975;2004).

    Parameters
    ----------
    dx : numpy array
        The proposed newton step.

    x : numpy array
        Parameter values for the equilibrium problem to be solved.

    endmembers_per_phase : list of integers
        A list of the number of endmembers in each phase.

    Returns
    -------
    lmda_bounds : tuple of floats
        minimum and maximum allowed fractions of the full newton step (dx).
    """

    max_steps = np.ones((len(x)))*100000.

    # first two constraints are P and T
    max_steps[0:2] = [20.e9, 500.]  # biggest reasonable P and T steps

    j = 2
    for i, n in enumerate(endmembers_per_phase):
        # if the phase fraction constraint would be broken,
        # set a step that is marginally smaller
        if x[j] + dx[j] < 0.:
            max_steps[j] = max(x[j]*0.999, 0.001)
        max_steps[j+1:j+n] = [max(xi*0.99, 0.01) for xi in x[j+1:j+n]]  # maximum compositional step
        j += n

    max_lmda = min([1. if step <= max_steps[i] else max_steps[i]/step
                    for i, step in enumerate(np.abs(dx))])

    return (1.e-8, max_lmda)


def phase_fraction_constraints(phase, assemblage, fractions, prm):
    """
    Converts a phase fraction constraint into standard linear form
    that can be processed by the root finding problem.

    Parameters
    ----------
    phase : burnman.SolidSolution or burnman.Mineral
        The phase for which the fraction is to be constrained

    assemblage : burnman.Composite
        The assemblage for which equilibrium is to be calculated.

    fractions : numpy array
        The phase fractions to be satified at equilibrium.

    prm : namedtuple
        A tuple with attributes n_parameters
        (the number of parameters for the current equilibrium problem)
        and phase_amount_indices (the indices of the parameters that
        correspond to phase amounts).

    Returns
    -------
    constraints : list
        An list of the phase fraction constraints.
    """
    phase_idx = assemblage.phases.index(phase)

    constraints = []
    for fraction in fractions:
        constraints.append(['X', [np.zeros((prm.n_parameters)), 0.]])
        constraints[-1][-1][0][prm.phase_amount_indices] = -fraction
        constraints[-1][-1][0][prm.phase_amount_indices[phase_idx]] += 1.

    return constraints


def phase_composition_constraints(phase, assemblage, constraints, prm):
    """
    Converts a phase composition constraint into standard linear form
    that can be processed by the root finding problem.

    We start with constraints in the form (site_names, n, d, v), where
    n*x/d*x = v and n and d are fixed vectors of site coefficients.
    So, one could for example choose a constraint
    ([Mg_A, Fe_A], [1., 0.], [1., 1.], [0.5]) which would
    correspond to equal amounts Mg and Fe on the A site.

    These are then converted by this function into endmember proportions
    (n'*p/d'*p = v). Because the proportions must add up to zero,
    we can reexpress this ratio into a linear constraint:
    [(n'[1:] - n'[0]) - v*(d'[1:] - d'[0])]*xi = v*d0 - n0
    which is less easy for a human to understand
    (in terms of chemical constraints), but easier to use as a constraint
    in a nonlinear solve.

    Parameters
    ----------
    phase : burnman.SolidSolution
        The phase for which the composition is to be constrained

    assemblage : burnman.Composite
        The assemblage for which equilibrium is to be calculated.

    constraints : a 4-tuple (list of strings, numpy array * 3)
        A tuple corresponding to the desired constraints, in the form
        (site_names, numerator, denominator, values).

    Returns
    -------
    x_constraints : list
        An list of the phase composition constraints in standard form.
    """
    phase_idx = assemblage.phases.index(phase)

    site_names, numerator, denominator, values = constraints
    site_indices = [phase.solution_model.site_names.index(name)
                    for name in site_names]
    noccs = phase.solution_model.endmember_noccupancies

    # Converts site constraints into endmember constraints
    # Ends up with shape (n_endmembers, 2)
    endmembers = np.dot(noccs[:, site_indices],
                        np.array([numerator, denominator]).T)

    numer0, denom0 = endmembers[0]
    endmembers -= endmembers[0]
    numer, denom = endmembers.T[:, 1:]

    # We start from the correct index
    start_idx = sum(assemblage.endmembers_per_phase[:phase_idx]) + 3
    n_indices = assemblage.endmembers_per_phase[phase_idx] - 1

    x_constraints = []
    for v in values:
        f = v*denom0 - numer0
        x_constraints.append(['X', [np.zeros((prm.n_parameters)), f]])
        x_constraints[-1][1][0][start_idx:start_idx+n_indices] = numer - v*denom

    return x_constraints


def get_equilibration_parameters(assemblage, composition):
    """
    Builds a named tuple containing the parameter names and
    various other parameters needed by the equilibrium solve.

    Parameters
    ----------
    assemblage : burnman.Composite
        The assemblage for which equilibrium is to be calculated.

    composition : dictionary
        The bulk composition for the equilibrium problem.

    Returns
    -------
    prm : namedtuple
        A tuple with attributes n_parameters
        (the number of parameters for the current equilibrium problem)
        and phase_amount_indices (the indices of the parameters that
        correspond to phase amounts).
    """
    # Initialize a named tuple for the equilibration parameters
    prm = namedtuple('assemblage_parameters', [])

    # Process parameter names
    prm.parameter_names = ['Pressure (Pa)', 'Temperature (K)']
    for i, n in enumerate(assemblage.endmembers_per_phase):
        prm.parameter_names.append('x({0})'.format(assemblage.phases[i].name))
        if n > 1:
            p_names = ['p({0} in {1})'.format(n, assemblage.phases[i].name)
                       for n in assemblage.phases[i].endmember_names[1:]]
            prm.parameter_names.extend(p_names)

    prm.n_parameters = len(prm.parameter_names)
    prm.phase_amount_indices = [i for i in range(len(prm.parameter_names))
                                if 'x(' in prm.parameter_names[i]]

    # Find the bulk composition vector
    prm.bulk_composition_vector = np.array([composition[e]
                                            for e in assemblage.elements])

    if assemblage.compositional_null_basis.shape[0] != 0:
        if (np.abs(assemblage.compositional_null_basis.dot(prm.bulk_composition_vector)[0]) > 1.e-12):
            raise Exception('The bulk composition is not within the '
                            'compositional space of the assemblage')

    prm.reduced_composition_vector = [prm.bulk_composition_vector[i]
                                      for i in assemblage.independent_element_indices]

    prm.constraint_vector, prm.constraint_matrix = calculate_constraints(assemblage)
    return prm


def process_eq_constraints(equality_constraints, assemblage, prm):
    """
    A function that processes the equality constraints
    into a form that can be processed by the F and jacobian functions.

    This function has two main tasks: it turns phase_fraction and
    phase_composition constraints into standard linear constraints in the
    solution parameters. It also turns vector-valued constraints into a
    list of scalar-valued constraints.

    Parameters
    ----------
    equality_constraints : list
        A list of equality constraints as provided by the user. For the
        types of constraints, please see the documentation for the
        equilibrate function.

    assemblage : burnman.Composite
        The assemblage for which equilibrium is to be calculated.

    prm : namedtuple
        A tuple with attributes n_parameters
        (the number of parameters for the current equilibrium problem)
        and phase_amount_indices (the indices of the parameters that
        correspond to phase amounts).

    Returns
    -------
    eq_constraint_lists : list of lists
        A list of lists of equality constraints in a form that can be processed
        by the F and jacobian functions.
    """
    eq_constraint_lists = []
    for i in range(2):
        if equality_constraints[i][0] == 'phase_fraction':
            phase = equality_constraints[i][1][0]
            fraction = equality_constraints[i][1][1]
            if isinstance(fraction, float):
                fraction = np.array([fraction])
            if not isinstance(fraction, np.ndarray):
                raise Exception('The constraint fraction in equality {0} '
                                'should be a float or numpy array'.format(i+1))
            eq_constraint_lists.append(phase_fraction_constraints(phase,
                                                                  assemblage,
                                                                  fraction,
                                                                  prm))
        elif equality_constraints[i][0] == 'phase_composition':
            phase = equality_constraints[i][1][0]
            constraint = equality_constraints[i][1][1]
            if isinstance(constraint[3], float):
                constraint = (constraint[0], constraint[1], constraint[2],
                              np.array([constraint[3]]))
            if not isinstance(constraint[3], np.ndarray):
                raise Exception('The last constraint parameter in equality '
                                '{0} should be a float '
                                'or numpy array'.format(i+1))

            eq_constraint_lists.append(phase_composition_constraints(phase,
                                                                     assemblage,
                                                                     constraint,
                                                                     prm))
        elif equality_constraints[i][0] == 'X':
            constraint = equality_constraints[i][1]
            if isinstance(constraint[-1], float):
                constraint = (constraint[0], np.array([constraint[-1]]))
            if not isinstance(constraint[-1], np.ndarray):
                raise Exception('The last constraint parameter in '
                                'equality {0} should be '
                                'a float or numpy array'.format(i+1))

            eq_constraint_lists.append([['X', [constraint[0], p]]
                                        for p in constraint[1]])

        elif (equality_constraints[i][0] == 'P'
              or equality_constraints[i][0] == 'T'
              or equality_constraints[i][0] == 'PT_ellipse'
              or equality_constraints[i][0] == 'S'
              or equality_constraints[i][0] == 'V'):
            if isinstance(equality_constraints[i][1], float):
                equality_constraints[i] = (equality_constraints[i][0],
                                           np.array([equality_constraints[i][1]]))
            if not isinstance(equality_constraints[i][1], np.ndarray):
                raise Exception('The last parameter in equality {0} should be '
                                'a float or numpy array'.format(i+1))
            eq_constraint_lists.append([[equality_constraints[i][0], p]
                                        for p in equality_constraints[i][1]])
        else:
            raise Exception('The type of equality_constraint is '
                            'not recognised for constraint {0}.\n'
                            'Should be one of P, T, S, V, X,\n'
                            'PT_ellipse, phase_fraction, '
                            'or phase_composition.'.format(i+1))
    return eq_constraint_lists


def equilibrate(composition, assemblage, equality_constraints,
                tol=1.e-3,
                store_iterates=False, store_assemblage=True,
                max_iterations=100., verbose=False):
    """
    A function that equilibrates an assemblage subject to given
    bulk composition and equality constraints by
    solving the equilibrium relations
    (chemical affinities for feasible reactions in the system
    should be equal to zero).

    Parameters
    ----------
    composition : dictionary
        The bulk composition that the assemblage must satisfy

    assemblage : burnman.Composite object
        The assemblage to be equilibrated

    equality_constraints : list
        A list of equality constraints. Each constraint
        should have the form: [<constraint type>, <constraint>], where
        <constraint type> is one of P, T, S, V, X, PT_ellipse,
        phase_fraction, or phase_composition. The <constraint> object should
        either be a float or an array of floats for P, T, S, V
        (representing the desired pressure, temperature,
        entropy or volume of the material). If the constraint type is X
        (a generic constraint on the solution vector) then the constraint c is
        represented by the following equality:
        np.dot(c[0], x) - c[1]. If the constraint type is
        PT_ellipse, the equality is given by
        norm(([P, T] - c[0])/c[1]) - 1.
        The constraint_type phase_fraction assumes a tuple of the phase object
        (which must be one of the phases in the burnman.Composite) and a float
        or vector corresponding to the phase fractions. Finally, a
        phase_composition constraint has the format (site_names, n, d, v),
        where n*x/d*x = v and n and d are fixed vectors of site coefficients.
        So, one could for example choose a constraint
        ([Mg_A, Fe_A], [1., 0.], [1., 1.], [0.5]) which would
        correspond to equal amounts Mg and Fe on the A site.

    tol : float
        The tolerance for the nonlinear solver.

    store_iterates : boolean
        Whether to store the parameter values for each iteration in
        each solution object.

    store_assemblage : boolean
        Whether to store a copy of the assemblage object in each
        solution object.

    max_iterations : integer
        The maximum number of iterations for the nonlinear solver.

    verbose : boolean
        Whether to print output updating the user on the status of
        equilibration.

    Returns
    -------
    sol_list : single, list, or 2D list of solver solution objects

    prm : namedtuple object
        A tuple with attributes n_parameters
        (the number of parameters for the current equilibrium problem)
        and phase_amount_indices (the indices of the parameters that
        correspond to phase amounts).
    """
    for ph in assemblage.phases:
        if isinstance(ph, SolidSolution) and not hasattr(ph, 'molar_fractions'):
            raise Exception(f'set_composition for solution {ph} before running equilibrate.')

    if assemblage.molar_fractions is None:
        n_phases = len(assemblage.phases)
        f = 1./float(n_phases)
        assemblage.set_fractions([f for i in range(n_phases)])

    assemblage.n_moles = (sum(composition.values())
                          / sum(assemblage.formula.values()))

    # Make parameter tuple
    prm = get_equilibration_parameters(assemblage, composition)

    # Check equality constraints have the correct structure
    # Convert into the format readable by the function and jacobian functions
    eq_constraint_lists = process_eq_constraints(equality_constraints,
                                                 assemblage, prm)

    # Set up solves
    sol_list = []
    n_c0 = len(eq_constraint_lists[0])
    n_c1 = len(eq_constraint_lists[1])

    # Find the initial state (could be none here)
    initial_state = [assemblage.pressure, assemblage.temperature]

    # Reset initial state if equality constraints
    # are related to pressure or temperature
    for i in range(2):
        if eq_constraint_lists[i][0][0] == 'P':
            initial_state[0] = eq_constraint_lists[i][0][1]
        elif eq_constraint_lists[i][0][0] == 'T':
            initial_state[1] = eq_constraint_lists[i][0][1]
        elif eq_constraint_lists[i][0][0] == 'PT_ellipse':
            initial_state = eq_constraint_lists[i][0][1][1]

    if initial_state[0] is None:
        initial_state[0] = 5.e9
    if initial_state[1] is None:
        initial_state[1] = 1200.

    assemblage.set_state(*initial_state)
    prm.initial_parameters = get_parameters(assemblage)
    parameters = get_parameters(assemblage)

    # Solve the system of equations, loop over input parameters
    sol_list = np.empty(shape=(n_c0, n_c1)+(0,)).tolist()
    for i_c0 in range(n_c0):
        new_c0 = True
        for i_c1 in range(n_c1):
            if verbose:
                string = 'Processing solution'
                if n_c0 > 1:
                    string += ' {0}/{1}'.format(i_c0+1, n_c0)
                if n_c1 > 1:
                    string += ' {0}/{1}'.format(i_c1+1, n_c1)
                print(string+':')

            equality_constraints = [eq_constraint_lists[0][i_c0],
                                    eq_constraint_lists[1][i_c1]]

            # Set the initial fractions and compositions
            # of the phases in the assemblage:
            # set_compositions_and_state_from_parameters(assemblage,
            #                                           prm.initial_parameters)
            try:
                sol = damped_newton_solve(F=lambda x: F(x, assemblage,
                                                        equality_constraints,
                                                        prm.reduced_composition_vector),
                                          J=lambda x: jacobian(x, assemblage,
                                                               equality_constraints),
                                          lambda_bounds=lambda dx, x: lambda_bounds(dx, x, assemblage.endmembers_per_phase),
                                          guess=parameters,
                                          linear_constraints=(prm.constraint_matrix,
                                                              prm.constraint_vector),
                                          tol=tol,
                                          store_iterates=store_iterates,
                                          max_iterations=max_iterations)
            except AttributeError:
                pass

            if store_assemblage:
                sol.assemblage = assemblage.copy()

            if verbose:
                print(sol.text)

            sol_list[i_c0][i_c1] = sol
            new_c0 = False

            prev = []
            if i_c1 < n_c1 - 1:
                next_ecs = [i_c0, i_c1 + 1]
            elif i_c1 == n_c1 - 1 and i_c0 < n_c0 - 1:
                next_ecs = [i_c0+1, 0]
            else:  # last value
                next_ecs = None

            if next_ecs is not None:
                cs = [eq_constraint_lists[0][next_ecs[0]],
                      eq_constraint_lists[1][next_ecs[1]]]
                prev_sol = []
                if next_ecs[0] != 0:
                    prev_sol.append(sol_list[next_ecs[0] - 1][next_ecs[1]])
                if next_ecs[1] != 0:
                    prev_sol.append(sol_list[next_ecs[0]][next_ecs[1] - 1])

                updated_params = False
                for s in prev_sol:
                    if s.success and not updated_params:
                        # next guess based on a Newton step
                        dF = F(s.x, assemblage, cs,
                               prm.reduced_composition_vector)
                        luJ = lu_factor(s.J)
                        new_parameters = s.x + lu_solve(luJ, -dF)
                        c = (prm.constraint_matrix.dot(new_parameters)
                             + prm.constraint_vector)
                        if all(c <= 0.):  # accept new guess
                            parameters = new_parameters
                        else:  # use the parameters from this step
                            parameters = s.x
                            exhausted_phases = [assemblage.phases[phase_idx].name
                                                for phase_idx, v in
                                                enumerate(new_parameters[prm.phase_amount_indices]) if v < 0.]
                            if len(exhausted_phases) > 0 and verbose:
                                print('A phase might be exhausted before the next step: {0}'.format(exhausted_phases))

                        updated_params = True
                if not updated_params:
                    parameters = prm.initial_parameters

    # Finally, make dimensions of sol_list equal the input dimensions
    if len(sol_list[0]) == 1:
        sol_list = list(zip(*sol_list))[0]
    if len(sol_list) == 1:
        sol_list = sol_list[0]
    return sol_list, prm



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for the Earth and Planetary Sciences
# Copyright (C) 2012 - 2017 by the BurnMan team, released under the GNU
# GPL v2 or later.


"""
Matas_etal_2007
^^^^^^^^^^^^^^^

Minerals from Matas et al. 2007 and references therein. See Table 1 and 2.
"""
from __future__ import absolute_import

from ..classes.mineral import Mineral


class mg_perovskite(Mineral):

    def __init__(self):
        self.params = {
            'equation_of_state': 'mgd2',
            'V_0': 24.43e-6,
            'K_0': 250.0e9,
            'Kprime_0': 4.0,
            'G_0': 175.0e9,
            'Gprime_0': 1.8,
            'molar_mass': .1020,
            'n': 5,
            'Debye_0': 1070.,
            'grueneisen_0': 1.48,
            'q_0': 1.4}

        Mineral.__init__(self)


class fe_perovskite(Mineral):

    def __init__(self):
        self.params = {
            'equation_of_state': 'mgd2',
            'V_0': 25.34e-6,
            'K_0': 250.0e9,
            'Kprime_0': 4.0,
            'G_0': 135.0e9,
            'Gprime_0': 1.3,
            'molar_mass': .1319,
            'n': 5,
            'Debye_0': 841.,
            'grueneisen_0': 1.48,
            'q_0': 1.4}

        Mineral.__init__(self)


class al_perovskite(Mineral):

    def __init__(self):
        self.params = {
            'equation_of_state': 'mgd2',
            'V_0': 24.58e-6,
            'K_0': 249.0e9,
            'Kprime_0': 4.0,
            'G_0': 165.0e9,
            'Gprime_0': 1.8,
            'molar_mass': .1005,
            'n': 5,
            'Debye_0': 1021.,
            'grueneisen_0': 1.48,
            'q_0': 1.4}

        Mineral.__init__(self)


class ca_perovskite(Mineral):

    def __init__(self):
        self.params = {
            'equation_of_state': 'mgd2',
            'V_0': 27.45e-6,
            'K_0': 236.0e9,
            'Kprime_0': 3.9,
            'G_0': 165.0e9,
            'Gprime_0': 2.46,
            'molar_mass': .11616,
            'n': 5,
            'Debye_0': 984.,
            'grueneisen_0': 1.53,
            'q_0': 1.6}

        Mineral.__init__(self)


class periclase (Mineral):

    def __init__(self):
        self.params = {
            'equation_of_state': 'mgd2',
            'V_0': 11.25e-6,
            'K_0': 160.1e9,
            'Kprime_0': 3.83,
            'G_0': 130.0e9,
            'Gprime_0': 2.2,
            'molar_mass': .0403,
            'n': 2,
            'Debye_0': 673.,
            'grueneisen_0': 1.41,
            'q_0': 1.3}

        Mineral.__init__(self)


class wuestite (Mineral):

    def __init__(self):
        self.params = {
            'equation_of_state': 'mgd2',
            'V_0': 12.26e-6,
            'K_0': 160.1e9,
            'Kprime_0': 3.83,
            'G_0': 46.0e9,
            'Gprime_0':  0.6,
            'molar_mass': .0718,
            'n': 2,
            'Debye_0': 673.,
            'grueneisen_0': 1.41,
            'q_0': 1.3}

        Mineral.__init__(self)

ca_bridgmanite = ca_perovskite
mg_bridgmanite = mg_perovskite
fe_bridgmanite = fe_perovskite
al_bridgmanite = al_perovskite



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for the Earth and Planetary Sciences
# Copyright (C) 2012 - 2017 by the BurnMan team, released under the GNU
# GPL v2 or later.


"""

HP_2011_fluids
^^^^^^^^^^^^^^

Fluids from Holland and Powell 2011 and references therein.
CORK parameters are taken from various sources.

CHO gases from Holland and Powell, 1991:
  - ["CO2",304.2,0.0738]
  - ["CH4",190.6,0.0460]
  - ["H2",41.2,0.0211]
  - ["CO",132.9,0.0350]

H2O and S2 from Wikipedia, 2012/10/23:
  - ["H2O",647.096,0.22060]
  - ["S2",1314.00,0.21000]

H2S from ancyclopedia.airliquide.com, 2012/10/23:
  - ["H2S",373.15,0.08937]

NB: Units for cork[i] in Holland and Powell datasets are:
  - a = kJ^2/kbar*K^(1/2)/mol^2: multiply by 1e-2
  - b = kJ/kbar/mol: multiply by 1e-5
  - c = kJ/kbar^1.5/mol: multiply by 1e-9
  - d = kJ/kbar^2/mol: multiply by 1e-13

Individual terms are divided through by P, P, P^1.5, P^2, so:
  - [0][j]: multiply by 1e6
  - [1][j]: multiply by 1e3
  - [2][j]: multiply by 1e3
  - [3][j]: multiply by 1e3
  - cork_P is given in kbar: multiply by 1e8
"""

from __future__ import absolute_import

from ..classes.mineral import Mineral
from ..tools.chemistry import dictionarize_formula, formula_mass


class CO2 (Mineral):

    def __init__(self):
        formula = 'CO2'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'carbon dioxide',
            'formula': formula,
            'equation_of_state': 'cork',
            'cork_params': [[5.45963e1, -8.63920e0], [9.18301e-1], [-3.30558e-2, 2.30524e-3], [6.93054e-4, -8.38293e-5]],
            'cork_T': 304.2,
            'cork_P': 0.0738e8,
            'H_0': -393.51e3,
            'S_0': 213.7,
            'Cp': [87.8, -2.644e-3, 706.4e3, -998.9],
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}
        Mineral.__init__(self)


class CH4 (Mineral):

    def __init__(self):
        formula = 'CH4'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'methane',
            'formula': formula,
            'equation_of_state': 'cork',
            'cork_params': [[5.45963e1, -8.63920e0], [9.18301e-1], [-3.30558e-2, 2.30524e-3], [6.93054e-4, -8.38293e-5]],
            'cork_T': 190.6,
            'cork_P': 0.0460e8,
            'H_0': -74.81e3,
            'S_0': 186.26,
            'Cp': [150.1, 0.002063, 3427700., -2650.4],
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}
        Mineral.__init__(self)


class O2 (Mineral):

    def __init__(self):
        formula = 'O2'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'oxygen',
            'formula': formula,
            'equation_of_state': 'cork',
            'cork_params': [[5.45963e1, -8.63920e0], [9.18301e-1], [-3.30558e-2, 2.30524e-3], [6.93054e-4, -8.38293e-5]],
            'cork_T': 0.,
            'cork_P': 1.0e5,
            'H_0': 0.,
            'S_0': 205.2,
            'Cp': [48.3, -0.000691, 499200., -420.7],
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}
        Mineral.__init__(self)


class H2 (Mineral):

    def __init__(self):
        formula = 'H2'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'hydrogen',
            'formula': formula,
            'equation_of_state': 'cork',
            'cork_params': [[5.45963e1, -8.63920e0], [9.18301e-1], [-3.30558e-2, 2.30524e-3], [6.93054e-4, -8.38293e-5]],
            'cork_T': 41.2,
            'cork_P': 0.0211e8,
            'H_0': 0.,
            'S_0': 130.7,
            'Cp': [23.3, 0.004627, 0.0, 76.3],
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}
        Mineral.__init__(self)


class S2 (Mineral):

    def __init__(self):
        formula = 'S2'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'sulfur',
            'formula': formula,
            'equation_of_state': 'cork',
            'cork_params': [[5.45963e1, -8.63920e0], [9.18301e-1], [-3.30558e-2, 2.30524e-3], [6.93054e-4, -8.38293e-5]],
            'cork_T': 1314.00,
            'cork_P': 0.21000e8,
            'H_0': 128.54e3,
            'S_0': 231.0,
            'Cp': [37.1, 0.002398, -161000.0, -65.0],
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}
        Mineral.__init__(self)


class H2S (Mineral):

    def __init__(self):
        formula = 'H2S'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'hydrogen sulfide',
            'formula': formula,
            'equation_of_state': 'cork',
            'cork_params': [[5.45963e1, -8.63920e0], [9.18301e-1], [-3.30558e-2, 2.30524e-3], [6.93054e-4, -8.38293e-5]],
            'cork_T': 373.15,
            'cork_P': 0.08937e8,
            'H_0': 128.54e3,
            'S_0': 231.0,
            'Cp': [47.4, 0.010240, 615900., -397.8],
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}
        Mineral.__init__(self)



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for
# the Earth and Planetary Sciences
# Copyright (C) 2012 - 2021 by the BurnMan team, released under the GNU
# GPL v2 or later.


from __future__ import absolute_import

import numpy as np
from sympy import Matrix
from ..classes.combinedmineral import CombinedMineral
from ..classes.solidsolution import SolidSolution
from .chemistry import site_occupancies_to_strings


def _decompose_3D_matrix(W):
    """
    Decomposes a 3D matrix W_ijk where E = W_ijk p_i p_j p_k
    into a subregular form where
    E = G_i p_i + WB_ij (1 - p_j + p_i) / 2 + WT_ijk p_i p_j p_k,
    and i < j < k.

    Parameters
    ----------
    W : 3D numpy array

    Returns
    -------
    new_endmember_excesses : 1D numpy array
        The array G_i

    new_binary_matrix : 2D numpy array
        The upper triangular matrix WB_ij

    new_ternary_terms : list of lists of length 4
        A list where each item is in the form [i, j, k, WT_ijk]
    """

    n_mbrs = len(W)
    # New endmember components
    # W_iii needs to be copied, otherwise just a view onto W
    new_endmember_excesses = np.copy(np.einsum('iii->i', W))

    # Removal of endmember components from 3D representation
    W -= (np.einsum('i, j, k->ijk',
                    new_endmember_excesses, np.ones(n_mbrs),
                    np.ones(n_mbrs))
          + np.einsum('i, j, k->ijk',
                      np.ones(n_mbrs), new_endmember_excesses,
                      np.ones(n_mbrs))
          + np.einsum('i, j, k->ijk',
                      np.ones(n_mbrs), np.ones(n_mbrs),
                      new_endmember_excesses))/3.

    # Transformed 2D components
    # (i=j, i=k, j=k)
    new_binary_matrix = (np.einsum('jki, jk -> ij', W, np.identity(n_mbrs))
                         + np.einsum('jik, jk -> ij', W, np.identity(n_mbrs))
                         + np.einsum('ijk, jk -> ij', W,
                                     np.identity(n_mbrs))).round(decimals=12)

    # Wb is the 3D matrix corresponding to the terms in the binary matrix,
    # such that the two following print statements produce the same answer
    # for a given array of endmember proportions
    Wb = (np.einsum('ijk, ij->ijk', W, np.identity(n_mbrs))
          + np.einsum('ijk, jk->ijk', W, np.identity(n_mbrs))
          + np.einsum('ijk, ik->ijk', W, np.identity(n_mbrs)))

    # Remove binary component from 3D representation
    # The extra terms are needed because the binary term in the formulation
    # of a subregular solution model given by
    # Helffrich and Wood includes ternary components (the sum_k X_k part)..
    W -= Wb + (np.einsum('ij, k', new_binary_matrix, np.ones(n_mbrs))
               - np.einsum('ij, ik->ijk',
                           new_binary_matrix, np.identity(n_mbrs))
               - np.einsum('ij, jk->ijk',
                           new_binary_matrix, np.identity(n_mbrs)))/2.

    # Find the 3D components Wijk by adding the elements at
    # the six equivalent positions in the matrix
    new_ternary_terms = []
    for i in range(n_mbrs):
        for j in range(i+1, n_mbrs):
            for k in range(j+1, n_mbrs):
                val = (W[i, j, k] + W[j, k, i]
                       + W[k, i, j] + W[k, j, i]
                       + W[j, i, k] + W[i, k, j]).round(decimals=12)
                if np.abs(val) > 1.e-12:
                    new_ternary_terms.append([i, j, k, val])

    return (new_endmember_excesses, new_binary_matrix, new_ternary_terms)


def _subregular_matrix_conversion(new_basis, binary_matrix,
                                  ternary_terms=None, endmember_excesses=None):
    """
    Converts the arrays reguired to describe a subregular solution
    from one endmember basis to another.

    The excess nonconfigurational energies of the subregular solution model
    are described as follows:
    E = G_i p_i + WB_ij (1 - p_j + p_i) / 2 + WT_ijk p_i p_j p_k,
    and i < j < k.

    Parameters
    ----------
    new_basis : 2D numpy array
        The new endmember basis, given as amounts of the old endmembers.

    binary_matrix : 2D numpy array
        The upper triangular matrix WB_ij

    ternary_terms : list of lists of length 4
        A list where each item is in the form [i, j, k, WT_ijk]

    endmember_excesses : 1D numpy array
        The array G_i

    Returns
    -------
    new_endmember_excesses : 1D numpy array
        The array G_i

    new_binary_matrix : 2D numpy array
        The upper triangular matrix WB_ij

    new_ternary_terms : list of lists of length 4
        A list where each item is in the form [i, j, k, WT_ijk]
    """
    n_mbrs = len(binary_matrix)
    # Compact 3D representation of original interactions
    W = (np.einsum('i, jk -> ijk', np.ones(n_mbrs), binary_matrix)
         + np.einsum('ij, jk -> ijk', binary_matrix, np.identity(n_mbrs))
         - np.einsum('ij, ik -> ijk', binary_matrix, np.identity(n_mbrs)))/2.

    # Add endmember components to 3D representation
    if endmember_excesses is not None:
        W += (np.einsum('i, j, k->ijk', endmember_excesses,
                        np.ones(n_mbrs), np.ones(n_mbrs))
              + np.einsum('j, i, k->ijk', endmember_excesses,
                          np.ones(n_mbrs), np.ones(n_mbrs))
              + np.einsum('k, i, j->ijk', endmember_excesses,
                          np.ones(n_mbrs), np.ones(n_mbrs)))/3.

    # Add ternary values to 3D representation
    if ternary_terms is not None:
        for i, j, k, val in ternary_terms:
            W[i, j, k] += val

    # Transformation to new 3D representation
    A = new_basis.T
    Wn = np.einsum('il, jm, kn, ijk -> lmn', A, A, A, W)

    new_endmember_excesses, new_binary_terms, new_ternary_terms = _decompose_3D_matrix(
        Wn)

    return (new_endmember_excesses, new_binary_terms, new_ternary_terms)


def complete_basis(basis):
    """
    Creates a full basis by filling remaining rows with
    rows of the identity matrix with row indices not
    in the column pivot list of the basis RREF
    """

    n, m = basis.shape
    if n < m:
        pivots = list(Matrix(basis).rref()[1])
        return np.concatenate((basis,
                               np.identity(m)[[i for i in range(m)
                                               if i not in pivots], :]),
                              axis=0)
    else:
        return basis


def transform_solution_to_new_basis(solution, new_basis, n_mbrs=None,
                                    solution_name=None, endmember_names=None,
                                    molar_fractions=None):
    """
    Transforms a solution model from one endmember basis to another.
    Returns a new SolidSolution object.

    Parameters
    ----------
    solution : :class:`burnman.SolidSolution` object
        The original solution object.

    new_basis : 2D numpy array
        The new endmember basis, given as amounts of the old endmembers.

    n_mbrs : float (optional)
        The number of endmembers in the new solution
        (defaults to the length of new_basis)

    solution_name : string (optional)
        A name corresponding to the new solution

    endmember_names : list of strings (optional)
        A list corresponding to the names of the new endmembers.

    molar_fractions : numpy array (optional)
        Fractions of the new endmembers in the new solution.

    Returns
    -------
    solution : :class:`burnman.SolidSolution` object
        The transformed solid solution
    """
    new_basis = np.array(new_basis)
    if n_mbrs is None:
        n_mbrs, n_all_mbrs = new_basis.shape
    else:
        _, n_all_mbrs = new_basis.shape

    if solution_name is None:
        name = 'child solution'
    else:
        name = solution_name

    solution_type = solution.solution_type
    if solution_type == 'ideal':
        ESV_modifiers = [[0., 0., 0.] for v in new_basis]

    elif (solution_type == 'asymmetric'
          or solution_type == 'symmetric'):

        A = complete_basis(new_basis).T

        old_alphas = solution.solution_model.alphas
        alphas = np.einsum('i, ij', solution.solution_model.alphas, A)
        inv_diag_alphas = np.diag(1./alphas)
        B = np.einsum('ij, jk, kl->il',
                      np.diag(old_alphas),
                      A, inv_diag_alphas)
        alphas = list(alphas[0:n_mbrs])
        Qe = np.einsum('ik, ij, kl->jl', solution.solution_model.We, B, B)
        Qs = np.einsum('ik, ij, kl->jl', solution.solution_model.Ws, B, B)
        Qv = np.einsum('ik, ij, kl->jl', solution.solution_model.Wv, B, B)

        def new_interactions(Q, n_mbrs):
            return [[float((Q[i, j] + Q[j, i] - Q[i, i] - Q[j, j])
                           * (alphas[i] + alphas[j])/2.)
                     for j in range(i+1, n_mbrs)]
                    for i in range(n_mbrs-1)]

        energy_interaction = new_interactions(Qe, n_mbrs)
        entropy_interaction = new_interactions(Qs, n_mbrs)
        volume_interaction = new_interactions(Qv, n_mbrs)

        ESV_modifiers = [[Qe[i, i] * alphas[i],
                          Qs[i, i] * alphas[i],
                          Qv[i, i] * alphas[i]]
                         for i in range(n_mbrs)]

    elif solution_type == 'subregular':
        full_basis = complete_basis(new_basis)

        def new_interactions(W, n_mbrs):
            return [[[W[i, j], W[j, i]] for j in range(i+1, n_mbrs)]
                    for i in range(n_mbrs-1)]

        # N.B. initial endmember_excesses are zero
        Emod, We, ternary_e = _subregular_matrix_conversion(full_basis,
                                                            solution.solution_model.We,
                                                            solution.solution_model.ternary_terms_e)
        Smod, Ws, ternary_s = _subregular_matrix_conversion(full_basis,
                                                            solution.solution_model.Ws,
                                                            solution.solution_model.ternary_terms_s)
        Vmod, Wv, ternary_v = _subregular_matrix_conversion(full_basis,
                                                            solution.solution_model.Wv,
                                                            solution.solution_model.ternary_terms_v)

        energy_interaction = new_interactions(We, n_mbrs)
        entropy_interaction = new_interactions(Ws, n_mbrs)
        volume_interaction = new_interactions(Wv, n_mbrs)

        ESV_modifiers = [[Emod[i], Smod[i], Vmod[i]] for i in range(n_mbrs)]

    else:
        raise Exception('The function to change basis for the '
                        '{0} solution model has not yet been '
                        'implemented.'.format(solution_type))

    # Create site formulae
    new_occupancies = np.array(new_basis).dot(
        solution.solution_model.endmember_occupancies)
    site_formulae = site_occupancies_to_strings(solution.solution_model.sites,
                                                solution.solution_model.site_multiplicities,
                                                new_occupancies)

    # Create endmembers
    endmembers = []
    for i, vector in enumerate(new_basis):
        nonzero_indices = np.nonzero(vector)[0]
        if len(nonzero_indices) == 1:
            endmembers.append([solution.endmembers[nonzero_indices[0]][0],
                               site_formulae[i]])
        else:
            mbr = CombinedMineral([solution.endmembers[idx][0]
                                   for idx in nonzero_indices],
                                  [vector[idx] for idx in nonzero_indices],
                                  ESV_modifiers[i])
            mbr.params['formula'] = {key: value for (key, value)
                                     in mbr.params['formula'].items()
                                     if value > 1.e-12}
            endmembers.append([mbr, site_formulae[i]])

    if endmember_names is not None:
        for i in range(n_mbrs):
            endmembers[i][0].params['name'] = endmember_names[i]
            endmembers[i][0].name = endmember_names[i]

    if n_mbrs == 1:
        endmembers[0][0].name = name
        endmembers[0][0].parent = solution
        endmembers[0][0].basis = new_basis
        return endmembers[0][0]
    else:
        new_solution = SolidSolution(name=name,
                                     solution_type=solution_type,
                                     endmembers=endmembers,
                                     energy_interaction=energy_interaction,
                                     volume_interaction=volume_interaction,
                                     entropy_interaction=entropy_interaction,
                                     alphas=alphas,
                                     molar_fractions=molar_fractions)
        new_solution.parent = solution
        new_solution.basis = new_basis
        return new_solution
